(* Chk_total_fifo *)
(* Georges Brun-Cottan *)
(* Based on Chk_fifo.ml and Chk_total *)
(* Tests for TOTAL *and* FIFO ordering properties on application casts *)
(**************************************************************)
open Layer
open Trans
open Util
open View
open Event
(**************************************************************)
let name = Trace.filel "CHK_APPL_FIFO"
(**************************************************************)

type from = Endpt.id
and id = int
and dest = Endpt.id
and origin = rank

type header = NoHdr
  | ACast of from * id * seqno
  | FCast of from * id * seqno
  | Gossip of rank Arrayf.t

(**************************************************************)

type s = {
  debug : string ;
  mutable cast_dn : seqno ;
  cast_up : seqno array ;
  send_dn : seqno array ;
  send_up : seqno array
} 

(**************************************************************)

let unique = counter ()

(**************************************************************)

let init _ (ls,vs) = {
  debug = "(???)" ;
  cast_dn = 0 ;
  cast_up = Array.create ls.nmembers 0 ;
  send_dn = Array.create ls.nmembers 0 ;
  send_up = Array.create ls.nmembers 0 ;
} 

let dump (ls,vs) s =
  eprintf "%s:dump:%s\n" name ls.name ;
  eprintf "  cast_dn=%d\n" s.cast_dn ;
  eprintf "  cast_up=%s\n" (string_of_int_array s.cast_up) ;
  eprintf "  send_dn=%s\n" (string_of_int_array s.send_dn) ;
  eprintf "  send_up=%s\n" (string_of_int_array s.send_up) 

let hdlrs s (ls,vs) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let failwith m = failwith (name^":"^s.debug^":"^m) in
  let log = Trace.log2 name ls.name in
  let recd = Queuee.create () in
  let got_view = ref false in
  let chk_ranks = ref None in
  let chk_rmt = ref [] in
  let chk_total () =
    match !chk_ranks with 
    | Some(ranks) ->
	let ranks = Arrayf.to_array ranks in
	List.iter 
	  (fun (rmt_origin,rmt_ranks) ->
	    let rmt_ranks = Arrayf.to_array rmt_ranks in
	    log (fun () -> sprintf "verifying:%d->%d" ls.rank rmt_origin) ;
	    if rmt_ranks <> ranks then 
	      begin
		eprintf "%s:TOTAL:ordering failure{%s}\n" name ls.name ;
		eprintf "  rank: %d rmt_rank: %d\n" ls.rank rmt_origin ;
		eprintf "  view_id=%s\n" (View.string_of_id ls.view_id) ;
		eprintf "     my cast:%s\n" (string_of_int_array ranks) ;
		eprintf "    rmt cast:%s\n" (string_of_int_array rmt_ranks) ;
		failwith "ordering failure"
	      end )
	  !chk_rmt;
	chk_rmt := []
    | None -> ()
  in
  let chk_fifo ev chk_from unique_id seqno =
    let origin = getPeer ev in
    let err1 = (seqno <> s.cast_up.(origin)) && (origin != ls.rank)
    and err2 = chk_from <> Arrayf.get vs.view origin in
    if err1 || err2 then
      begin
        if err1 then
	  begin
            eprintf "%s:%s:multicast message out of order unique_id=%d (origin=%d, my_rank=%d, seqno=%d, expect=%d)\n"
	      name s.debug unique_id origin ls.rank seqno s.cast_up.(origin)
	  end ;
	
        if err2 then 
	  begin
            if err1 then eprintf "%s:***AND***\n" name ;
            eprintf "%s:incorrect origin m.mendpt, sent=(%s) <> (%s,%d)=my_info\n"
	      name (Endpt.string_of_id chk_from)  (Endpt.string_of_id (Arrayf.get vs.view origin)) origin
	  end ;
      end ;
    array_incr s.cast_up origin ;
  in

  let up_hdlr ev abv hdr = match getType ev,hdr with
  | ECast, FCast(chk_from,unique_id,seqno) ->
      if !got_view then
      	failwith "ECast after EView" ;
      chk_fifo ev chk_from unique_id seqno;
      up ev abv

  | ECast, ACast(chk_from,unique_id,seqno) ->
      if !got_view then
      	failwith "ECast after EView" ;
      Queuee.add (getPeer ev) recd ;
      chk_fifo ev chk_from unique_id seqno;
      up ev abv

  | _, NoHdr ->
      up ev abv

  | _ -> failwith "bad up event"


  and uplm_hdlr ev hdr = match getType ev,hdr with
  | ECast,Gossip(ranks)  ->
      let origin = getPeer ev in
      chk_rmt := (origin,ranks) :: !chk_rmt ;
      chk_total () ;
      free_noIov name ev
  | _ -> failwith "unknown local message"
	
  and upnm_hdlr ev = match getType ev with  
  | EView ->
      got_view := true ;
      let ranks = Queuee.to_list recd in
      let ranks = Arrayf.of_list ranks in
      chk_ranks := Some ranks ;
      chk_total () ;
      dnlm (create name ECast[NoTotal]) (Gossip ranks) ;
      upnm ev
  | EDump -> dump (ls,vs) s ; upnm ev
  | _ -> upnm ev

  and dn_hdlr ev abv = match (getType ev, getApplMsg ev) with
  | (ECast, true) ->
      dn ev abv 
	(if getNoTotal ev then 
	  (FCast(ls.endpt,(unique()),s.cast_dn)) 
	else
	  (ACast(ls.endpt,(unique()),s.cast_dn)) );
      s.cast_dn <- succ s.cast_dn

  | _ -> dn ev abv NoHdr

  and dnnm_hdlr ev = dnnm ev

  in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}

let l args vf = Layer.hdr init hdlrs None NoOpt args vf

let _ =   Elink.layer_install name l

(**************************************************************)
