(**************************************************************)
(* Fifo.ml : FIFO reliable multicast protocol *)
(* Added Fifo delivery property to application's *)
(*                  'NoTotal Appl' ECast *)
(* ESend events are still not ordered wrt. to ECast *)
(* Author: Georges Brun-Cottan, 1/1999 *)
(* Based on MNAK code by: Mark Hayden *)
(**************************************************************)
open Ensemble
open View
open Event
open Util
open Layer
open Trans
open Iq
(**************************************************************)
let name = Trace.filel "Fifo"
(**************************************************************)

type header = NoHdr
  | NoTotal  of seqno
  | Total    of seqno

(**************************************************************)
(* Guaranteed properties: 
   * - delivery order of any two multicast originating from 
   *   different sources is kept intact.
   * - delivery order of any two multicast originating from 
   *   the same source is FIFO.
   *
   * The FIFO Layer's goal is to add FIFO property between Total and 
   * NoTotal multicast from one source
   *)

type 'abv state = 
    {
    mutable seqno  : seqno;
    buf    : 'abv Iq.t Arraye.t;	(* Fifo queues *)
    queue  : (rank * seqno) Queue.t	(* Total queue *)
    }

(**************************************************************)

let dump (ls,vs) s =
  eprintf "Fifo:dump:%s\n" ls.name ;
  eprintf "  rank=%d, nmembers=%d\n" ls.rank ls.nmembers ;
  (let len = (Queue.length s.queue) in 
  let queueState = 
    if len = 0 then 
      "(None)"
    else
      let (rank,seqno) =  Queue.peek s.queue in 
      sprintf "(rank= %d, seqno= %d)" rank seqno
  in 
  eprintf "  queue=%d  top:%s\n" len queueState);
  eprintf "  cast_lo=%s\n" (Arraye.int_to_string (Arraye.map Iq.lo s.buf)) ;
  eprintf "  cast_hi=%s\n" (Arraye.int_to_string (Arraye.map Iq.hi s.buf)) ;
  eprintf "  read   =%s\n" (Arraye.int_to_string (Arraye.map Iq.read s.buf))

(**************************************************************)

let init _ (ls,vs) = 
  let buf = Arraye.init ls.nmembers (fun _ -> Iq.create name Local_nohdr) in
  let iq_init = Param.int vs.params "mnak_iq_init" in
  if iq_init > 0 then
    Arraye.iter (fun iq -> Iq.grow iq iq_init) buf ;
  { seqno= (Iq.read (Arraye.get buf ls.rank)); 
    buf = buf; 
    queue=Queue.create()}

(**************************************************************)
exception NotOrdered

let hdlrs s (ls,vs) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let log = Trace.log2 name ls.name in
  let logb = Trace.log2 "BUFFER" (name^":"^ls.name) in
  let name = "Fifo:"^ls.name in
  let failwith m = dump (ls,vs) s ; failwith (name^":"^m) in

  let fastFifoOrderedP = Iq.opt_insert_check in 
  let totalP buf seqno = ((Queue.length s.queue) = 0) in 
  let checkForImmediateDelivery buf seqno = 
    (fastFifoOrderedP buf seqno) & (totalP buf seqno) in

  let read_prefix read_prefix_fun rank buf =
    read_prefix_fun
      buf 
      (fun seqno iov abv ->
	log (fun () -> sprintf "read_prefix:%d:Data(%d)" rank seqno);
	let iov = Iovecl.copy name iov in
	up (castPeerIov name rank iov) abv)
  in
  let recv_cast rank seqno abv iov =
    let buf = Arraye.get s.buf rank in

    (* We first FIFO enqueue *)
    (ignore (Iq.assign buf seqno iov abv));
    log (fun () -> sprintf "recv_cast:%d:assign=%d" rank seqno);
    
    (* For each total Ecast in the queue we first deliver fifo Ecast *) 
    (try 
      Queue.iter 
	(fun (origin,seqno) -> 
	  let buf = Arraye.get s.buf origin in
	  if read_prefix (Iq.read_bounded_prefix seqno) origin buf then
	    begin
	      (ignore (Queue.take s.queue));
	      ()
	    end
	  else raise NotOrdered)
	s.queue
    with
      NotOrdered -> ());
    
    (* we delivered all total ECast. Check for possible FIFO pending ECast *) 
    if ((Queue.length s.queue) = 0) then
      Arraye.iteri (read_prefix Iq.read_prefix) s.buf
  in

  let up_hdlr ev abv hdr = 
    match getType ev, hdr with

    | (ECast, Total(seqno)) ->
	let origin = getPeer ev in
	let buf = Arraye.get s.buf origin in
	let iov = getIov ev in

         (* Check for Fast-path *)
	 (* means the total multicast is in Fifo order *)
	if ((checkForImmediateDelivery buf seqno) || (origin = ls.rank)) then
	  (* I shortcut when the message come from myself because
	     we assume that localfifo layer is not below. 
	     It still should work if it there because the stack should guarantee the fifo
	     proper delivery, so this is a win-win test. well... *) 
	  begin
	    (ignore (Iq.opt_insert_doread buf seqno iov abv));
	    Iq.set_lo buf seqno; (* should not be needed *)
	    up ev abv
	  end
	else
	  begin
	    Queue.add (origin,seqno) s.queue;
	    recv_cast origin seqno abv iov;
	    Event.free name ev
	  end

    | (ECast, NoTotal(seqno)) ->
	let origin = getPeer ev in
	let buf = Arraye.get s.buf origin in
	let iov = getIov ev in

         (* Check for Fast-path *)
	 (* means the total multicast is in Fifo order *)
	if checkForImmediateDelivery buf seqno then
	  begin
	    (ignore (Iq.opt_insert_doread buf seqno iov abv));
	    up ev abv
	  end
	else
	  begin
	    recv_cast origin seqno abv iov;
            Event.free name ev
	  end

    | _, NoHdr -> up ev abv
    | _        -> failwith "bad header"

  and uplm_hdlr ev hdr = failwith "unknown local message"
  and upnm_hdlr ev = 
    match getType ev with
    | EExit ->
         (* GC all buffers. *)
	Arraye.iter Iq.free s.buf ;
	upnm ev

    | EAccount ->
         (* Dump information about status of buffers. *)
	logb (fun () -> sprintf "msgs=%s"
            (Arraye.int_to_string (Arraye.map (fun c -> Iq.hi c - Iq.lo c) s.buf))) ;
	upnm ev

    | EDump -> ( dump (ls,vs) s ; upnm ev )
    | _ -> upnm ev

  and dn_hdlr ev abv =
    match (getType ev, getApplMsg ev) with
     (* ECast: buffer a copy of all application ECast and send it on. *)
    | (ECast,true) ->
	let seqno = s.seqno in
	s.seqno <- (succ seqno);
	dn ev abv
	  (if getNoTotal ev then 
	    (NoTotal seqno)
	  else
	    (Total seqno))

    | _ -> dn ev abv NoHdr

  and dnnm_hdlr = dnnm

  in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}

let l args vs = Layer.hdr init hdlrs None NoOpt args vs

let _ = 
  Param.default "mnak_iq_init" (Param.Int 0) ;
  Elink.layer_install name l
