(**************************************************************)
(* Fifo.ml : FIFO reliable multicast protocol *)
(* Added Fifo delivery property to application's *)
(*                  'NoTotal Appl' ECast *)
(* ESend events are ordered wrt. to ECast *)
(* Author: Georges Brun-Cottan, 6/1999 *)
(* Based on MNAK code by: Mark Hayden *)
(**************************************************************)
open Ensemble
open View
open Event
open Util
open Layer
open Trans
(**************************************************************)
let name = Trace.filel "FifoCS"
(**************************************************************)

module MultipartTimeStamp = struct 
  type seqno = int
  type s = seqno array
	
  let init_state nmem v = Array.create nmem v

   (* Increment our own counter *)
  let incr t org = t.(org) <- succ t.(org)

   (* Increment all counters *)
  let incr_all t = Array.iteri (fun org _ -> incr t org) t

  let sprint t = 
    Printf.sprintf "[%s]"
      (String.concat "," 
	 (Array.to_list (Array.map (fun x -> string_of_int x) t)))
end

type header = | NoHdr | TS of MultipartTimeStamp.s

(**************************************************************)
(* Guaranteed properties: 
   * - delivery order of any pair of multicast and send originating from 
   *   the same source is FIFO.
   *)

type 'abv state = 
    {
    ts     : MultipartTimeStamp.s; (* send clock*) 
    buf    : 'abv Iq.t array; 

    (* For debugging *) 
(* 
    recv_send :  MultipartTimeStamp.s;
    recv_cast :  MultipartTimeStamp.s;
    max_send :  MultipartTimeStamp.s;
    max_cast :  MultipartTimeStamp.s;
    deliver_send :  MultipartTimeStamp.s;
    deliver_cast :  MultipartTimeStamp.s;
    direct_deliver_send :  MultipartTimeStamp.s;
    direct_deliver_cast :  MultipartTimeStamp.s;
*)
    }

(**************************************************************)

let dump (ls,vs) s =
  let stringify f a =
    let g buf = string_of_int (f buf) in 
    String.concat " " (Array.to_list (Array.map g s.buf))
  in
  eprintf "FifoCS:dump:%s\n" ls.name ;
  eprintf "  rank=%d, nmembers=%d send_clock=%s\n" 
    ls.rank ls.nmembers (MultipartTimeStamp.sprint s.ts);
  eprintf "  cast_lo=%s\n" (stringify Iq.lo s.buf);
  eprintf "  cast_hi=%s\n" (stringify Iq.hi s.buf) ;
  eprintf "  read   =%s\n" (stringify Iq.read s.buf)

(*
  eprintf "  recv_send=%s\n"  (MultipartTimeStamp.sprint s.recv_send);
  eprintf "  recv_cast=%s\n"  (MultipartTimeStamp.sprint s.recv_cast);
  eprintf "  max_send=%s\n"  (MultipartTimeStamp.sprint s.max_send);
  eprintf "  max_cast=%s\n"  (MultipartTimeStamp.sprint s.max_cast);
  eprintf "  deliver_send=%s\n"  (MultipartTimeStamp.sprint s.deliver_send);
  eprintf "  deliver_cast=%s\n"  (MultipartTimeStamp.sprint s.deliver_cast);
  eprintf "  direct_deliver_send=%s\n"  (MultipartTimeStamp.sprint s.direct_deliver_send);
  eprintf "  direct_deliver_cast=%s\n"  (MultipartTimeStamp.sprint s.direct_deliver_cast)
*)


let init _ (ls,vs) = 
  let buf = Array.init ls.nmembers (fun _ -> Iq.create name (Local_nohdr,EPrompt)) in
  let iq_init = Param.int vs.params "fifocs_iq_init" in
  let first_seqno = (Iq.read (Array.get buf ls.rank)) in
  if iq_init > 0 then
    begin
      Array.iter (fun iq -> Iq.grow iq iq_init) buf
    end;
  
  { ts = MultipartTimeStamp.init_state ls.nmembers first_seqno;
    buf = buf; 
    
(*  recv_send = MultipartTimeStamp.init_state ls.nmembers first_seqno;
    recv_cast = MultipartTimeStamp.init_state ls.nmembers first_seqno;
    max_send = MultipartTimeStamp.init_state ls.nmembers 0;
    max_cast = MultipartTimeStamp.init_state ls.nmembers 0;
    deliver_send = MultipartTimeStamp.init_state ls.nmembers first_seqno;
    deliver_cast = MultipartTimeStamp.init_state ls.nmembers first_seqno;
    direct_deliver_send = MultipartTimeStamp.init_state ls.nmembers first_seqno;
    direct_deliver_cast = MultipartTimeStamp.init_state ls.nmembers first_seqno;
*)  
}

let hdlrs s (ls,vs) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let log = Trace.log2 name ls.name in
  let logb = Trace.log2 "BUFFER" (name^":"^ls.name) in
  let name = "FifoCS:"^ls.name in
  let failwith m = 
    dump (ls,vs) s; 
    failwith (name^":"^m) 
  in
  let checkForImmediateDelivery buf seqno =
    Iq.opt_insert_check buf seqno
  in 

  let recv_cast rank seqno evt iov =
    log (fun () -> sprintf "recv_cast:%d:assign=%d" rank seqno);
    let buf = Array.get s.buf rank in
    let read_prefix rank buf = 
      let upcall seqno iov (abv, event_type) = 
	log (fun () -> sprintf "read_prefix:%d:Data(%d)" rank seqno);
	let iov = Iovecl.copy name iov in
	match event_type with 
	| ECast ->
	    (* MultipartTimeStamp.incr s.deliver_cast rank; *)
	    up (castPeerIov name rank iov) abv
	| ESend -> 
	    (* MultipartTimeStamp.incr s.deliver_send rank; *)
	    up (sendPeerIov name rank iov) abv
	| _ -> failwith "Corrupted Queue" 
      in
      Iq.read_prefix buf upcall
    in
    ignore (Iq.assign buf seqno iov evt);
    Array.iteri read_prefix s.buf
  in

  let up_hdlr ev abv hdr = 
    match getType ev, hdr with
    | ((ECast|ESend), TS(ts)) ->
	let event_type = (abv, (getType ev)) in
	let seqno = Array.get ts ls.rank in 
	let origin = getPeer ev in
	let buf = Array.get s.buf origin in
	let iov = getIov ev in

	(* 
	log (fun () -> 
	  sprintf "receive:%d:data(%d):%s" 
	    origin seqno
	    (match  (getType ev) with 
	    | ECast -> "ECast"
	    | ESend -> "ESend") );
	   *)
	(* Debug *)
	(* 
	(match (getType ev) with
	|  ECast -> 
	    s.max_cast.(origin) <- max s.max_cast.(origin) seqno;
	    MultipartTimeStamp.incr s.recv_cast origin;
	|  ESend -> 
	    s.max_send.(origin) <- max s.max_send.(origin) seqno;
	    MultipartTimeStamp.incr s.recv_send origin);
	*)
         (* Check for Fast-path  (Multicast is already in Fifo order) *)
	if (checkForImmediateDelivery buf seqno)  then
	  begin
	    let _ = Iq.opt_insert_doread buf seqno iov event_type in
(*
	    (match (getType ev) with
	    |  ECast -> MultipartTimeStamp.incr s.direct_deliver_cast origin
	    |  ESend -> MultipartTimeStamp.incr s.direct_deliver_send origin);
*)
	    up ev abv
	  end
	else
	  begin
	    recv_cast origin seqno event_type iov;
	    free name ev
	  end

    | _, NoHdr -> up ev abv
    | _        -> failwith "bad header"

  and uplm_hdlr ev hdr = failwith "unknown local message"
  and upnm_hdlr ev = 
    match getType ev with
    | EBlock -> 
	 (* Check for sanity. No messages should still be in buffers *)
	Array.iter
	  (fun iq -> 
	    if (Iq.read iq) < (Iq.hi iq) then 
	      dump (ls,vs) s)
	  s.buf;
	upnm ev

    | EExit ->
	 (* Check for sanity. No messages should still be in buffers *)
	Array.iter
	  (fun iq -> assert ((Iq.read iq) = (Iq.hi iq)) )
	  s.buf;
	
         (* Free all buffers. *)
	Array.iter Iq.free s.buf ;
	upnm ev

    | ETimer -> 
         (* GC all buffers. *)
	Array.iter 
	  (fun iq -> Iq.set_lo iq (Iq.read iq))
	  s.buf;
	upnm ev

    | EDump -> 
	dump (ls,vs) s;
	upnm ev
    | _ -> upnm ev

  and dn_hdlr ev abv =
    match (getType ev, getApplMsg ev) with
    | (ECast, true) ->
	let ts = Array.copy s.ts in
	MultipartTimeStamp.incr_all s.ts;
	dn ev abv (TS ts)
    | (ESend, true) ->
	let ts = Array.copy s.ts in
	MultipartTimeStamp.incr s.ts (getPeer ev);
	dn ev abv (TS ts)
    | _ -> dn ev abv NoHdr
	  
  and dnnm_hdlr = dnnm

  in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}

let l args vs = Layer.hdr init hdlrs None NoOpt args vs

let _ =  
  Param.default "fifocs_iq_init" (Param.Int 0) ;
  Elink.layer_install name l
