(**************************************************************)
(* ping.ml : user-level ping protocol *)
(* Author: Georges Brun-Cottan, 6/1999 *)
(**************************************************************)
open View
open Event
open Util
open Layer
open Trans

open Unix
(**************************************************************)
let name = Trace.filel "Ping"
(**************************************************************)

type state = { 
    ping_period       : Time.t;
    mutable next_ping : Time.t;
    mutable rcvd_ping : int;
    mutable ts        : float;
    mutable logs       : (float list) array;
  } 
      
(**************************************************************)

let init _ (ls,vs) = {
  ping_period = Param.time vs.params "ping_period";
  next_ping = Time.zero;
  rcvd_ping = 0;
  ts = 0.0;
  logs = Array.init ls.nmembers (fun _ -> []); 
}
    
let hdlrs s (ls,vs) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let notping = ( ls.nmembers < (Param.int vs.params "ping_quorum")) in 
  let failwith m = failwith (name^":"^m) in
  let dump_logs () = 
     (* Dump the ping logs into the logfile *)
    let dump_log f msg log = 
      let moments l = 
	let bi_sum (m1,m2) n = (m1 +. n, m2 +. (n *. n)) in 
	List.fold_left bi_sum (0.0, 0.0) l
      in
      let n = (List.length log) in 
      let (m1,m2) = moments log in 
      let mean = m1 /. (float n) in
      let var = m2 /. (float n) -. mean *. mean in
      Printf.fprintf f "# %s [n= %d, mean= %f, variance= %f]\n" msg n mean var;
      List.iter (fun t -> Printf.fprintf f "%f\n" t) log;
      Printf.fprintf f "\n\n"; (* New data set -- gnuplot convention *)
    in
 
    let log_name = Param.string vs.params "ping_log" in
    let f = open_out_gen [Open_wronly;Open_creat;Open_append] 0o666 log_name in
    Util.eprintf "\nDumping ping stats in %s\n" log_name;
    Printf.fprintf f "#### %s ####\n" (View.string_of_id ls.view_id);
    Array.iteri
      (fun rank l ->
	if rank <> ls.rank then 
	  let msg = 
	    Printf.sprintf "# Ping RTT %s <--> %s (ping rate= %s)" 
	      (Endpt.string_of_id ls.endpt) 
	      (Endpt.string_of_id (Arrayf.get vs.view rank))
	      (Time.to_string s.ping_period)
	  in
	  dump_log f msg (List.rev l) )
      s.logs; 
    close_out f
  in

  let got_all_pings () = (s.rcvd_ping <= 1) in
  let send_pings () = 
    s.ts <- Unix.gettimeofday();
    s.rcvd_ping <- ls.nmembers;
    (* dnlm (castEv name) () *)
    dnlm (create name ECast [ApplMsg]) ()
  in
  let recv_ping peer = 
     (* Got back my ping *)
    let time = Unix.gettimeofday() in 
    s.logs.(peer) <- (time -. s.ts) :: s.logs.(peer);
    s.rcvd_ping <- pred s.rcvd_ping;
  in

  let up_hdlr ev abv () = up ev abv
  and uplm_hdlr ev () =
    let peer = getPeer ev in
    match getType ev with
    | ECast -> 
         (* Bounce the ping *)
	dnlm (create name ESend [Peer peer; ApplMsg]) ();
	 (* dnlm (sendPeer name peer) (); *)
	free name ev
    | ESend -> 
	recv_ping peer;
	free name ev
    | _ -> failwith "Unknown local message"
	
  and upnm_hdlr ev = 
    if notping || (Time.is_zero s.ping_period) then  
      upnm ev
    else
      match getType ev with
      | EView ->
	  dump_logs();
	  upnm ev
      | ETimer -> 
	  let time = getTime ev in
	  if (got_all_pings()) && (Time.Ord.ge time s.next_ping) then 
	    begin
	      s.next_ping <- Time.add time s.ping_period;
	      send_pings()
	    end;
	  upnm ev
      | _ -> upnm ev

  and dn_hdlr ev abv = dn ev abv ()
  and dnnm_hdlr = dnnm

  in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}

let l args vs = Layer.hdr init hdlrs None NoOpt args vs

let _ =  
  Param.default "ping_period" (Param.Time (Time.of_int 1));
  Param.default "ping_log" (Param.String "pingLog");
  Param.default "ping_quorum" (Param.Int 2);
  Elink.layer_install name l





