(* pingPiggy.ml : logs an aproximation of the transport rtt of the
   application msgs *)

(* This ping is specificaly designed for the LESS divergence control
   protocol when used with the Bench program.  In particular, it
   assume that Casts and Acks are fifo and that every Casts is
   acknowledged exactly once in the order it has been delivered.  *)

(**************************************************************)
open View
open Event
open Util
open Layer
open Trans

open Unix
(**************************************************************)
let name = Trace.filel "PingPiggy"
(**************************************************************)

type header = NoHdr | TS | Delta of float

type state = {
    mutable logs  : (float list) array;
    mutable send_casts : seqno;
    mutable send_acks  : seqno;
    recv_acks     : seqno array;
    recv_casts    : seqno array;
    casts         : (seqno, float * (bool array)) Hashtbl.t
    acks          : (seqno, float) Hashtbl.t
  } 
      
(**************************************************************)

let init _ (ls,vs) = {
  logs  = Array.create ls.nmembers [];
  send_casts = 0;
  send_acks  = 0;
  recv_acks  = Array.create ls.nmembers 0;
  recv_casts = Array.create ls.nmembers 0;
  casts      =  Hashtbl.create 23459;
  acks       =  Hashtbl.create 23459;
}


let hdlrs s (ls,vs) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let failwith m = failwith (name^":"^m) in
  let acked = Array.init ls.nmembers (fun i -> (i = ls.rank)) in 

   (* Emitter parts *)
  let send_cast () =
    Hashtbl.add s.casts s.send_casts (Unix.gettimeofday(), Array.copy acked);
    s.send_casts <- succ s.send_casts;
    TS
  in
  let recv_ack origin delta = 
    let for_all a =
      let rec loop i =
	if i < 0 then true
	else if a.(i) then loop (pred i)
	else false
      in loop (pred (Array.length a))
    in
    let t2 = Unix.gettimeofday() in
    let seq = s.recv_acks.(origin) in 
    let (t1, acked) = Hashtbl.find s.casts seq in
    acked.(origin) := true;
    if for_all acked then Hashtbl.remove s.casts seq;
    s.recv_acks.(origin) = succ seq;
    s.logs.(origin) <- (t2 -. (t1 +. delta)) :: s.logs.(origin);
  in

   (* Receiver parts *)
  let recv_cast origin =
    let seq = s.recv_casts.(origin) in
    Hashtbl.add s.acks s.seq (Unix.gettimeofday());
    s.recv_casts.(origin) = succ seq
  in
  let send_ack origin =
    let seq = s.recv_casts.(origin) in
    let t2 = Unix.gettimeofday() in
    let t1 = Hashtbl.find s.acks seq in
    Hashtbl.remove s.acks seq;
    Delta (t2 -. t1)
  in

  let dump_logs () = 
     (* Dump the ping logs into the logfile *)
    let dump_log f log = 
      let moments l = 
	let bi_sum (m1,m2) n = (m1 +. n, m2 +. (n *. n)) in 
	List.fold_left bi_sum (0.0, 0.0) l
      in
      let n = (List.length log) in 
      if n > 0 then 
	begin
	  let (m1,m2) = moments log in 
	  let mean = m1 /. (float n) in
	  let var = m2 /. (float n) -. mean *. mean in
	  Printf.fprintf f "# [n= %d, mean= %f, variance= %f]\n" n mean var;
	  List.iter (fun t -> Printf.fprintf f "%f\n" t) log;
	  Printf.fprintf f "\n\n"	(* New data set -- gnuplot convention *)
      	end
      else
	Printf.fprintf f "# No data logged for ping\n\n\n"
    in
    
    if (List.length log) > 0 then 
      begin
	let log_name = Param.string vs.params "ping_log" in
	let f = open_out_gen [Open_wronly;Open_creat;Open_append] 0o666 log_name in
	Util.eprintf "\nDumping ping stats in %s\n" log_name;
	Printf.fprintf f "#### %s ####\n" (View.string_of_id ls.view_id);
	Array.iteri
	  (fun rank l ->
	    if rank <> ls.rank then 
	      let msg = 
		Printf.sprintf "# Ping RTT %s <--> %s" 
		  (Endpt.string_of_id ls.endpt) 
		  (Endpt.string_of_id (Arrayf.get vs.view rank))
	      in
	      dump_log f (List.rev l) )
	  s.logs; 
	close_out f
      end
  in

  let up_hdlr ev abv hdr =
    match getType ev, hdr with
    | ECast, PingRequest -> 
         (* Bounce the ping *)
	dnlm (Event.create name ESend [Peer (getPeer ev); ApplMsg]) PingReply;
	up ev abv
    | _ -> up ev abv

  and uplm_hdlr ev hdr =
    match getType ev, hdr with
    | ESend, PingReply -> 
	recv_ping (getPeer ev);
	free name ev
    | _ -> failwith "Unknown local message"
	  
  and upnm_hdlr ev = 
    if notping || (Time.is_zero s.ping_period) then  
      upnm ev
    else
      match getType ev with
      | EView ->
	  dump_logs();
	  upnm ev
      | ETimer -> 
	  let time = getTime ev in
	  if (got_all_pings()) && (Time.Ord.ge time s.next_ping) then 
	    begin
	      s.next_ping <- Time.add time s.ping_period;
	      send_ping := true;
	    end;
	  upnm ev
      | _ -> upnm ev

  and dn_hdlr ev abv = 
    if notping || (Time.is_zero s.ping_period) then
      dn ev abv NoHdr
    else
      if (getApplMsg ev) && !send_ping then
	send_pings ev abv
      else
	dn ev abv NoHdr

  and dnnm_hdlr = dnnm

  in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}

let l args vs = Layer.hdr init hdlrs None NoOpt args vs

let _ =  
  Param.default "ping_log" (Param.String "pingLog");
  Elink.layer_install name l
