(* Stlatency -- 'Sample' latency --  Simulate high-latency communication *) 

(* Distributed under the FSF LGPL license *)
(* Copyright Georges Brun-Cottan <bruncott@cs.brandeis.edu> *)
(* Brandeis University Jan 1999*) 

(* We postpone every (ECast|ESend) upwards deliveries for a given
   delay (latency_value) (in milliseconds).  
   *)

(* We enqueue every message. Every ETimer event we check for
   delayed message. ETimer acts as the sample rate. It is assumed to be set 
   by the user.

   If the reception is uniform in time, this give us a mean latency of
   'delay'.  See other latency layers for alternate distribution functions.
   *)

(* Use 'Param.default "latency_value" (Param.Time  (Time <your delay>))' to 
   configure the delay.  Default delay is zero.  There is 
   no overhead in the case delay is zero.
   
   The delay can not be changed. It is used at the time the stack is build.
   *)
open Ensemble
open View
open Event
open Util
open Layer
open Trans

(**************************************************************)
let name = Trace.filel "Stlatency"
(**************************************************************) 
type header = NoHdr | Timer
  
type ('abv) upEventKind = LEvent of Event.t * ('abv option)

type 'abv state = {
    delay         : Time.t; (* The delay parameterized by the user *)
    queue         : 'abv upEventKind Queue.t;
  } 

(**************************************************************)

let dump (ls,vs) s =
  let older = 
    if (Queue.length s.queue) = 0 then 
      (Time.of_int 0)
    else
      let LEvent(ev,abv) = Queue.peek s.queue in 
      (getTime ev)
  in
  eprintf "%s:dump: latency= %s\n" name  (Time.to_string s.delay);
  eprintf "\tNb of delayed messages= %d, older delayed message= %s\n" 
    (Queue.length s.queue) (Time.to_string older)

let debug = false

(**************************************************************)

exception NoMoreDelivery

let init _ (ls,vs) = 
  let log = Trace.log2 name ls.name in
  let delay = Param.time vs.params "latency_value" in 
  log (fun () -> sprintf "Latency = %s" (Time.to_string delay));
  { delay= delay;
    queue= Queue.create() } 
    

let hdlrs s (ls,vs) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let failwith m = failwith (name^":"^m) in 
  let alarm =  Elink.alarm_get_hack () in
  let delayMessage lev =
    let LEvent(ev,hdr) = lev in 
    let t = Alarm.gettime alarm in
    Queue.add (LEvent ((Event.set name ev [Time t]),hdr)) s.queue
  in
  let checkForDelivery time =
    try 
      Queue.iter 
	(function LEvent(ev,hdr) -> 
	  if Time.ge (Time.add (getTime ev) s.delay) time then
	    raise NoMoreDelivery
	  else 
	    begin
	      match hdr with 
	      | Some(abv) -> up ev abv
	      | None -> upnm ev;
	    end;
	  ignore (Queue.take s.queue); 
	  ())
	s.queue;
    with NoMoreDelivery -> ()
  in

  let up_hdlr = 
    if Time.is_zero s.delay then 
      (* Fast Path *)
      (fun ev abv hdr ->  up ev abv)
    else
      (* Slow Path *)
      (fun ev abv hdr ->
	match getType ev with 
	| (ECast|ESend) -> 
	    delayMessage (LEvent(ev,Some abv))
	| _ ->
	    up ev abv)
  and uplm_hdlr ev hdr =
    failwith "Unknown local message"
  and upnm_hdlr = 
    if Time.is_zero s.delay then 
      (* Fast Path *)
      (fun ev -> upnm ev)
    else
      (* Slow Path *)
      (fun ev ->
	match getType ev with 
	| (ECast|ESend) ->
	    delayMessage (LEvent(ev, None))
	| ETimer ->
	    checkForDelivery (getTime ev);
	    upnm ev
	| EDump -> 
	    dump (ls,vs) s;
	    upnm ev
	| _ ->
	    upnm ev)
  and dn_hdlr ev abv = dn ev abv NoHdr
  and dnnm_hdlr = dnnm
      
in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}

let l args vs = Layer.hdr init hdlrs None NoOpt args vs

let _ =
  Param.default "latency_value" (Param.Time (Time.of_int 0));
  Elink.layer_install name l

(**************************************************************)


