(* The Epsilon Divergence protocol.

   The protocol considers two kind of operations, those whose metrics
   is zero and the ones whose metrics is above zero.

   The protocol orders the execution of operations whose metrics is
   strictly positive according to the Less divergence protocol.  It
   causally orders (partially) the others: null metrics operations are
   executed after all the previous positive metrics operation have
   been executed. Previous means operation that have been executed
   before an operation.  Also, all operations are fifo ordered.

   Althouth we use general term design, this protocol has been devised
   explicitely for the LESS semaphore. In this exemple, TestAndDec
   operations are assigned a strictly positive metrics whereas Inc
   operations are assigned a null metrics. In this context, this
   protocol is proven to guarantee that the local replica state is a
   lower bound of the global state when a TestAndDec operation is to
   be *executed* on its initiation site 
   *)

(* Join protocol.

   Join requires the protocol state to support view change and to
   allow consistent initialisation of newcommers including replica and
   protocol state transfer as well as threshold negociation to allow
   newcommers to initiate operations.

   The protocol state consists into a queues of pending operations and
   the a vector clock.

   Persistancy of this state between view change is required, at least
   when queus are not empty at the time of the view change.  This
   happens in two situations: (i) when operations scheduled by the
   protocol are postponed by the application. This behavior depends
   from the application semantics but as it is the most general, it is
   the one implemented by this module. (ii) When there is pending
   operations initiated. This is application independant.

   The protocol uses members id in three places: (1) to associate
   operation requests (local or remote) to members (pending queues),
   (2) to associate acknowledgement to members (windows management)
   and (3) to enforce causal dependancies between members.

   (1) and (2) are addressed in translating the view dependant member
   id identifying the sender in a view independant one.  These
   translation upon receipt of every message. It is done through an
   associative array updated upon installation of a new view.

   (3) require no special care if the view is not a state transfer
   view (xfer). Total ordering of views allows to enforce causality
   between causal vector belonging to two distinct views. A state
   transfer view identifies a newcommer (one at a time) and require
   special treatment on the coordinator and the newcommer.
   Initialisation state is given to the newcommer by the
   coordinator. This includes the replica state, its pending queues
   and its vector clock.  Strong VS enforce the needed atomic property
   of this transfer.

   The initialisation protocol requires one and only one member to be
   a primary.  The coordinators of the succeeding views in which did
   belong this member will inherit the primary property.  The primary
   is used (1) to enable the protocol and (2) to decide who is the
   primary for state initialisation purpose.

   The protocol keep track of the primary property. When a non
   singleton new view is installed, the coordinator of the previous
   primary view is in charge of initialisation. It may be not the
   coordinator. Once initialised, the coordinator of the next non
   transfer view will act as the primary.  This protocol ensure a
   total order on primary views, starting from the application member
   stamped as primary.

   Members only have to keep track is they where primary coordinator
   of a previous view.

   The protocol behavior is unpredictable if two members are given
   concurrently the primary property -- Members have to agree on who
   is a primary.  A session manager healing active session members may
   be used to enforce this agreement.

   The threshold negociation starts once the initialisation is
   terminated. We provide three policies. The two first one are
   intended for member independant bound.  In the first policy,
   newcommers receive the coordinator threshold. In the second, they
   receive a locally specified threshold. In the last policy, the
   coordinator computes the desired thresholds for each members and
   send them to all members.  Members that are above this threshold
   are require not to initiate more operation until they passed below.
   Furthermore, they send back to the coordinator the threshold in
   excess they release. The coordinator dispatch incrementally this
   threshold to newcommer.

 *)

open Protocol_intf
open Replica_intf
open Measure_intf

module Div : Protocol =      
  functor (A : ADT) ->
    functor (MI : Metrics_intf) ->
      functor (M : Measure_intf with type op_t = A.op_t and type metrics = MI.t) -> 
  struct
    open Ensemble
    open View
    open Trans
    open Appl_intf open New

    open Causalts
    open VSCausal
    open Printf
    open Name

    exception Pending_request	

    type op_t =  A.op_t
    type metrics = MI.t
    type acker  = op_t -> unit

    type seqno = int
    type operationInstance = OperationInstance of op_t * seqno
    type causalprop = VSCausal.s option
    type operationMsg = 
      | DC of operationInstance * causalprop
      | Fifo of operationInstance * causalprop
      | Ack of seqno
      |	NewState of A.state * (operationMsg Queue.t) array * Endpt.id array

    type operation = operationMsg
    type issuer = (op_t * acker) -> (operation naction) list 

    type handlers = { 
	deliver : origin -> cast_or_send -> operation -> (operation naction) list;
        issue   : (op_t * acker) -> (operation naction) list; 
	dump    : unit -> unit;
	stable  : unit -> bool;
    }
  
    type operationStatus = {
	mutable initiated : int;
	mutable delivered : int;
	mutable scheduled : int;
      }

    let memberId_map = ref (Name.init_map ())

     (* There is only two way to be initialised. Be the primary or
	having been initialised by a primary *)

    let initialised = ref false	

    let install ((ls,vs) as vf) primary nmembers threshold = 

      let transfer_intf () = 	
	let dump () = Util.eprintf "Sorry no dump during state transfer" in
	let stable () = false in
	let fetch () = [] in 
	let issue _ = [] in 
	let deliver origin cs msg =
	  match cs with 
	  | C -> 
	      (match msg with
	      | NewState (newState, newQueues, translationVector) ->
		  A.install_state newState;
		  []
	      | _ -> failwith "deliver (xfer): invalid cast")
		
	  | S -> failwith "deliver (xfer): invalid send"
	in
	let handlers = {
	  dump = dump;
	  deliver = deliver;
	  issue = issue;
	  stable = stable}
	in
	([],handlers)
      in

      let running_intf () = 	
	 (* Monitoring *)
	let dcOperation = {initiated = 0; delivered = 0; scheduled = 0} in  
	let fifoOperation = {initiated = 0; delivered = 0; scheduled = 0} in  
	let receivedAck = ref 0 in 

	let timeStamp = Util.counter () in (* The local time stamp generator *)
	let getMyId = Name.get_fastmap !memberId_map vs.view in
	let string_of_id n = sprintf "[%d:%d]" n (getMyId n) in
	
       (* Vector clock keeping track of the causal DC relationship over execution order  *)
	let siteTime = VSCausal.create_state vf in

       (* Pending operations *)
	let fifoQueues = Array.init ls.nmembers (fun _ -> Queue.create ()) in 

       (* Not acknowledged operations initiated by this member. *)
	let nackedOps = Hashtbl.create 113 in 
	
       (* Metrics of not acknowledged operations *)
	let nackedThreshold = Array.create nmembers (MI.null()) in 

       (* Compute the maximum of current divergence amongst every member. *)
	let maxNackedThreshold () = 
	  let maxThreshold = ref (MI.null()) in 
	  Array.iter 
	    (fun t -> if (MI.gt t !maxThreshold) then maxThreshold := t)
	    nackedThreshold;
	  !maxThreshold
	in
	
	let notAboveTreshold () = not (MI.gt (maxNackedThreshold()) threshold) in

	let dump () =
	  let dump_threshold () = 
	    Util.eprintf "threshold : [";
	    Array.iter
	      (fun t -> Util.eprintf "%s " (MI.to_string t))
	      nackedThreshold;
	    Util.eprintf "]\n";
	  in
	  Util.eprintf "siteTime= %s\n" (VSCausal.sprint siteTime);
	  Util.eprintf "DC {I= %d, D= %d, S= %d}  Fifo {I= %d, D= %d, S= %d}  Ack=%d\n"
	    dcOperation.initiated dcOperation.delivered dcOperation.scheduled 
	    fifoOperation.initiated fifoOperation.delivered fifoOperation.scheduled 
	    !receivedAck;
	  dump_threshold ();
	  Array.iteri 
	    (fun i q -> 
	      Util.eprintf "Fifo %d (%d), top ts=%s\n"
		i (Queue.length q)
		(if (Queue.length q) > 0 then 
		  match Queue.peek q with 
		  | DC ((OperationInstance (op,ts) as opi),cs_opt) ->
		      sprintf "%s %d %s %s" 
			(A.to_string op) ts "(DC)"
			(match cs_opt with 
			| Some(cs) -> (VSCausal.sprint cs)
			| None ->
			    if ls.rank = i then 
			      begin
				"notAboveThreshold= " ^ 
				(if notAboveTreshold () then
				  "true"
				else
				  "false")
			      end
			    else
			      "No causal dependancy" )
		  | Fifo (OperationInstance (op,ts), cs_opt) ->
		      sprintf "%s %d %s %s" (A.to_string op) ts "(Fifo)"
			(match cs_opt with 
			| Some(cs) -> (VSCausal.sprint cs)
			| None -> "No causal dependancy" )
		  | Ack (ts) ->
		      sprintf "%d" ts
		  | NewState _ -> "New state"
		else 
		  "Empty" ))
	    fifoQueues;
	  Util.eprintf "\n"
	in

	let ackOp origin ts =
	  let for_all a =
	    let rec loop i =
	      if i < 0 then true
	      else if a.(i) then loop (pred i)
	      else false
	    in loop (pred (Array.length a))
	  in
	  try 
	    let (op, localAcker, acked) = Hashtbl.find nackedOps ts in  
	    acked.(origin) <- true;
	    nackedThreshold.(origin) <- (MI.sub nackedThreshold.(origin) (M.mu op));
	    let expectRemoteAck  = 
	      if origin = ls.rank then
		localAcker op
	      else
		true
	    in
	    if (not expectRemoteAck) || (for_all acked) then 
	      begin	 
		Hashtbl.remove nackedOps ts
	      end
	  with Not_found ->
	    failwith  
	      (Printf.sprintf "ackOp(%d:%d): Acknowledgement of a not referenced operation"
		 origin ts)
	in

	let pushOp exec_acker (OperationInstance (op, ts)) =
	  Hashtbl.add nackedOps ts (op, exec_acker, Array.create nmembers false);
	  for i = 0 to (nmembers - 1) do
	    nackedThreshold.(i) <- (MI.add nackedThreshold.(i) (M.mu op))
	  done
	in
	
     (* The core of the delivery algorithm.
	We iterate forever on fifoqueues until no more operations can be
	scheduled.  Returns a list of operations to broadcast *)

	let rec processQueues () =
	  let opToSend = ref [] in 
	  let scheduled = ref true in 

	  let ackop_l (OperationInstance (_,ts)) = ackOp ls.rank ts in

	  let scheduleTop q (OperationInstance(op,_)) = 
	    match (A.execute op) with 
	    | true -> 
		Queue.take q;
		scheduled := true;		(* keep track of the state change *)
		true
	    | false -> false
	  in

	  while !scheduled do		(* Iterate until no more scheduling is possible *)
	    scheduled := false;
	    Array.iteri
	      (fun orig q -> 
		let rec loop_sched q = 
		  if Queue.length q > 0 then 
		    begin
		      let op = Queue.peek q in
		      match op with 
		      | DC (opi,None) ->
			  assert (orig = ls.rank);
			  if (notAboveTreshold ()) && (scheduleTop q opi) then 
			    
			     (* Operations are always enqueued. We
				must check for the threshold criterion
				before initiating one.  Only now can
				we compute the causal dependancies and
				send them to others along with the
				operation *)

			    begin
			      let cs = VSCausal.copy siteTime in 
			      let OperationInstance(op,ts) = opi in 
			      VSCausal.incr siteTime orig;
			      dcOperation.scheduled <- succ dcOperation.scheduled;
			      ackop_l opi;
			      (Cast(DC(opi, Some(cs)))) :: (loop_sched q)
			    end
			  else
			    []

		      |  DC (OperationInstance(op,ts) as opi, Some(cs)) ->
			  assert (orig <> ls.rank);
			  if (VSCausal.ge siteTime cs) && (scheduleTop q opi) then
			    begin
			      VSCausal.incr siteTime orig;
			      dcOperation.scheduled <- succ dcOperation.scheduled;
			      (Send([|orig|], Ack(ts))) :: (loop_sched q)
			    end
			  else 
			    []

		      | Fifo (opi,None) ->
			  assert (orig = ls.rank);
			  if scheduleTop q opi then 
			    begin
			      ackop_l opi;
			      fifoOperation.scheduled <- succ fifoOperation.scheduled;

 			      (* Only now can we compute the causal dependancies 
				 and send them to others along with the operation *)

			      (Cast(Fifo(opi, Some(VSCausal.copy siteTime)))) :: (loop_sched q)
			    end
			  else
			    []

		      | Fifo (opi,Some(cs)) ->
			  assert (orig <> ls.rank);
			  if (VSCausal.ge siteTime cs) && (scheduleTop q opi) then
			    begin
			      fifoOperation.scheduled <- succ fifoOperation.scheduled;
			      loop_sched q
			    end
			  else 
			    []

		      | Ack (ts) -> 
			  let _ = Queue.take q in 
			  ackOp orig ts;
			  scheduled := true;
			  []	
			    
		      | NewState _ -> failwith "Late state transfer" 
		    end
		  else 
		    []
		in
		opToSend := !opToSend @ (loop_sched q) )
	      fifoQueues;	
	  done;
	  !opToSend

	and initOperation ((op, exec_acker) as request) =

	   (* The core algorithm for initiating operations.  We
 	      enqueue systematically.  `processQueue' check for the
 	      threshold criterion on the local queue before initiating
 	      an operation.
	      
	      The nacktreshold is first incremented/updated. The
 	      operation is initiated when the max of nacktreshold is
 	      below or equal the threshold protocol parameter 

	      The windows being incremented before the initiation, this
	      implementation assumes synchronous mode of operations
	      *)
	  
	  let opi = OperationInstance(op, timeStamp()) in
	  if MI.gt (M.mu op) (MI.null ()) then
 	     (* Non null metrics operation *)
	    begin
	      pushOp (fun op -> exec_acker op; true) opi;
	      dcOperation.initiated <- succ dcOperation.initiated;
	      Queue.add (DC (opi,None)) fifoQueues.(ls.rank)
	    end
	  else
	     (* Null metrics operation *)
	    begin
	      pushOp (fun op -> exec_acker op; false) opi;
	      fifoOperation.initiated <- succ fifoOperation.initiated;
	      Queue.add (Fifo (opi,None)) fifoQueues.(ls.rank);
	    end;
	  processQueues ()

	and issue request =

	  (* The protocol does not support asynchronous request.  This
	     is because the windows is updated right now at issuing
	     time not initiating time. This is transparent for the
	     user.  The needed change is small but not completely
	     obvious to be correct in the case where operation
	     execution can be postpone after thay have been scheduled.
	     Also, I guess that postponing execution of operation is
	     useful only for synchrounous mode. Also, I am not sure
	     about the semantics of divergence control together with
	     asynchronous operation. *)
	  
	  if Queue.length fifoQueues.(ls.rank) > 0 then 
	    begin
	      dump();
	      raise Pending_request
	    end;
	  initOperation request 
	in 
	
	(* Queue may be persistant. We tag operation by their deliver view. *)
	let deliver origin cs msg =
	  (match cs with 
	  | C -> 
	      (match msg with
	      | Fifo (_) -> 
		  fifoOperation.delivered <- succ fifoOperation.delivered
	      | DC (_) -> 
	       (* check_fifo_delivery origin msg; *)
		  dcOperation.delivered <- succ dcOperation.delivered
	      |	_ -> failwith "deliver: bad cast\n");
	      Queue.add msg fifoQueues.(origin)
	  | S -> 
	      (match msg with
	      | Ack (_) -> 
		  Pervasives.incr receivedAck;
		  Queue.add msg fifoQueues.(origin)
	      |	_ -> failwith "deliver: bad send\n") );
	  
       (* check_fifo_sendings (processQueues()) *)
	  processQueues()
	in
	let stable () = not (MI.gt (maxNackedThreshold()) (MI.null())) in 
	let handlers = {
	  dump = dump;
	  deliver = deliver;
	  issue = issue;
	  stable = stable}
	in
	([],handlers)
      in
      
      memberId_map := Name.update_map !memberId_map vs.view;
      if primary && (not !initialised) then 
	begin
	  assert (vs.ltime = 0);	(* we must be a new born singleton *)
	  initialised := true;
	end;
      match vs.xfer_view with 
      | true -> transfer_intf ()
      | false -> running_intf ()

  end
 
