(* The same as div protocol but do not enforce causality *)

open Protocol_intf
open Replica_intf
open Measure_intf

module Div_fifo : Protocol =      
  functor (A : ADT) ->
    functor (MI : Metrics_intf) ->
      functor (M : Measure_intf with type op_t = A.op_t and type metrics = MI.t) -> 
  struct
    open Ensemble
    open View
    open Trans
    open Appl_intf open New

    open Causalts
    open VSCausal
    open Printf
    open Name

    exception Pending_request	

    type op_t =  A.op_t
    type metrics = MI.t
    type acker  = op_t -> unit

    type seqno = int
    type operationInstance = OperationInstance of op_t * seqno
    type operationMsg = 
      | DC of operationInstance
      | Fifo of operationInstance
      | Ack of seqno
      |	NewState of A.state * (operationMsg Queue.t) array * Endpt.id array

    type operation = operationMsg
    type issuer = (op_t * acker) -> (operation naction) list 

    type handlers = { 
	deliver : origin -> cast_or_send -> operation -> (operation naction) list;
        issue   : (op_t * acker) -> (operation naction) list; 
	dump    : unit -> unit;
	stable  : unit -> bool;
    }
  
    type operationStatus = {
	mutable initiated : int;
	mutable delivered : int;
	mutable scheduled : int;
      }

    let memberId_map = ref (Name.init_map ())

     (* There is only two way to be initialised. Be the primary or
	having been initialised by a primary *)

    let initialised = ref false	

    let install ((ls,vs) as vf) primary nmembers threshold = 

      let transfer_intf () = 	
	let dump () = Util.eprintf "Sorry no dump during state transfer" in
	let stable () = false in
	let fetch () = [] in 
	let issue _ = [] in 
	let deliver origin cs msg =
	  match cs with 
	  | C -> 
	      (match msg with
	      | NewState (newState, newQueues, translationVector) ->
		  A.install_state newState;
		  []
	      | _ -> failwith "deliver (xfer): invalid cast")
		
	  | S -> failwith "deliver (xfer): invalid send"
	in
	let handlers = {
	  dump = dump;
	  deliver = deliver;
	  issue = issue;
	  stable = stable}
	in
	([],handlers)
      in

      let running_intf () = 	
	 (* Monitoring *)
	let dcOperation = {initiated = 0; delivered = 0; scheduled = 0} in  
	let fifoOperation = {initiated = 0; delivered = 0; scheduled = 0} in  
	let receivedAck = ref 0 in 

	let timeStamp = Util.counter () in (* The local time stamp generator *)
	let getMyId = Name.get_fastmap !memberId_map vs.view in
	let string_of_id n = sprintf "[%d:%d]" n (getMyId n) in

       (* Pending operations *)
	let fifoQueues = Array.init ls.nmembers (fun _ -> Queue.create ()) in 

       (* Not acknowledged operations initiated by this member. *)
	let nackedOps = Hashtbl.create 113 in 
	
       (* Metrics of not acknowledged operations *)
	let nackedThreshold = Array.create nmembers (MI.null()) in 

       (* Compute the maximum of current divergence amongst every member. *)
	let maxNackedThreshold () = 
	  let maxThreshold = ref (MI.null()) in 
	  Array.iter 
	    (fun t -> if (MI.gt t !maxThreshold) then maxThreshold := t)
	    nackedThreshold;
	  !maxThreshold
	in
	
	let notAboveTreshold () = not (MI.gt (maxNackedThreshold()) threshold) in

	let dump () =
	  let dump_threshold () = 
	    Util.eprintf "threshold : [";
	    Array.iter
	      (fun t -> Util.eprintf "%s " (MI.to_string t))
	      nackedThreshold;
	    Util.eprintf "]\n";
	  in
	  Util.eprintf "DC {I= %d, D= %d, S= %d}  Fifo {I= %d, D= %d, S= %d}  Ack=%d\n"
	    dcOperation.initiated dcOperation.delivered dcOperation.scheduled 
	    fifoOperation.initiated fifoOperation.delivered fifoOperation.scheduled 
	    !receivedAck;
	  dump_threshold ();
	  Array.iteri 
	    (fun i q -> 
	      Util.eprintf "Fifo %d (%d), top ts=%s\n"
		i (Queue.length q)
		(if (Queue.length q) > 0 then 
		  match Queue.peek q with 
		  | DC (OperationInstance (op,ts) as opi) ->
		      sprintf "%s %d %s %s" 
			(A.to_string op) ts "(DC)"
			(if ls.rank = i then 
			  begin
			    "notAboveThreshold= " ^ 
			    (if notAboveTreshold () then
			      "true"
			    else
			      "false")
			  end
			else "")
		  | Fifo (OperationInstance (op,ts)) ->
		      sprintf "%s %d %s" (A.to_string op) ts "(Fifo)"
		  | Ack (ts) ->
		      sprintf "%d" ts
		  | NewState _ -> "New state"
		else 
		  "Empty" ))
	    fifoQueues;
	  Util.eprintf "\n"
	in

	let ackOp origin ts =
	  let for_all a =
	    let rec loop i =
	      if i < 0 then true
	      else if a.(i) then loop (pred i)
	      else false
	    in loop (pred (Array.length a))
	  in
	  try 
	    let (op, localAcker, acked) = Hashtbl.find nackedOps ts in  
	    acked.(origin) <- true;
	    nackedThreshold.(origin) <- (MI.sub nackedThreshold.(origin) (M.mu op));
	    let expectRemoteAck  = 
	      if origin = ls.rank then
		localAcker op
	      else
		true
	    in
	    if (not expectRemoteAck) || (for_all acked) then 
	      begin	 
		Hashtbl.remove nackedOps ts
	      end
	  with Not_found ->
	    failwith  
	      (Printf.sprintf "ackOp(%d:%d): Acknowledgement of a not referenced operation"
		 origin ts)
	in

	let pushOp exec_acker (OperationInstance (op, ts)) =
	  Hashtbl.add nackedOps ts (op, exec_acker, Array.create nmembers false);
	  for i = 0 to (nmembers - 1) do
	    nackedThreshold.(i) <- (MI.add nackedThreshold.(i) (M.mu op))
	  done
	in
	
     (* The core of the delivery algorithm.
	We iterate forever on fifoqueues until no more operations can be
	scheduled.  Returns a list of operations to broadcast *)

	let rec processQueues () =
	  let opToSend = ref [] in 
	  let scheduled = ref true in 

	  let ackop_l (OperationInstance (_,ts)) = ackOp ls.rank ts in

	  let scheduleTop q (OperationInstance(op,_)) = 
	    match (A.execute op) with 
	    | true -> 
		Queue.take q;
		scheduled := true;		(* keep track of the state change *)
		true
	    | false -> false
	  in

	  while !scheduled do		(* Iterate until no more scheduling is possible *)
	    scheduled := false;
	    Array.iteri
	      (fun orig q -> 
		let rec loop_sched q = 
		  if Queue.length q > 0 then 
		    begin
		      let op = Queue.peek q in
		      match op with 
		      | DC (OperationInstance(op,ts) as opi) ->
			  if orig = ls.rank then
			    if (notAboveTreshold ()) && (scheduleTop q opi) then 
			    
			     (* Operations are always enqueued. We
				must check for the threshold criterion
				before initiating one.  Only now can
				we compute the causal dependancies and
				send them to others along with the
				operation *)

			      begin
				let OperationInstance(op,ts) = opi in 
				dcOperation.scheduled <- succ dcOperation.scheduled;
				ackop_l opi;
				(Cast (DC opi)) :: (loop_sched q)
			      end
			    else
			      []
			  else 
			    if scheduleTop q opi then
			      begin
				dcOperation.scheduled <- succ dcOperation.scheduled;
				(Send([|orig|], Ack(ts))) :: (loop_sched q)
			      end
			    else 
			      []
				
		      | Fifo opi ->
			  if orig = ls.rank then
			    if scheduleTop q opi then 
			      begin
				ackop_l opi;
				fifoOperation.scheduled <- succ fifoOperation.scheduled;
				
 			      (* Only now can we compute the causal dependancies 
				 and send them to others along with the operation *)
				
				(Cast(Fifo opi)) :: (loop_sched q)
			      end
			    else
			      []
			  else
			    if scheduleTop q opi then
			      begin
				fifoOperation.scheduled <- succ fifoOperation.scheduled;
				loop_sched q
			      end
			    else 
			      []
				
		      | Ack (ts) -> 
			  let _ = Queue.take q in 
			  ackOp orig ts;
			  scheduled := true;
			  []	
			    
		      | NewState _ -> failwith "Late state transfer" 
		    end
		  else 
		    []
		in
		opToSend := !opToSend @ (loop_sched q) )
	      fifoQueues;	
	  done;
	  !opToSend

	and initOperation ((op, exec_acker) as request) =

	   (* The core algorithm for initiating operations.  We
 	      enqueue systematically.  `processQueue' check for the
 	      threshold criterion on the local queue before initiating
 	      an operation.
	      
	      The nacktreshold is first incremented/updated. The
 	      operation is initiated when the max of nacktreshold is
 	      below or equal the threshold protocol parameter 

	      The windows being incremented before the initiation, this
	      implementation assumes synchronous mode of operations
	      *)
	  
	  let opi = OperationInstance(op, timeStamp()) in
	  if MI.gt (M.mu op) (MI.null ()) then
 	     (* Non null metrics operation *)
	    begin
	      pushOp (fun op -> exec_acker op; true) opi;
	      dcOperation.initiated <- succ dcOperation.initiated;
	      Queue.add (DC opi) fifoQueues.(ls.rank)
	    end
	  else
	     (* Null metrics operation *)
	    begin
	      pushOp (fun op -> exec_acker op; false) opi;
	      fifoOperation.initiated <- succ fifoOperation.initiated;
	      Queue.add (Fifo opi) fifoQueues.(ls.rank);
	    end;
	  processQueues ()

	and issue request =

	  (* The protocol does not support asynchronous request.  This
	     is because the windows is updated right now at issuing
	     time not initiating time. This is transparent for the
	     user.  The needed change is small but not completely
	     obvious to be correct in the case where operation
	     execution can be postpone after thay have been scheduled.
	     Also, I guess that postponing execution of operation is
	     useful only for synchrounous mode. Also, I am not sure
	     about the semantics of divergence control together with
	     asynchronous operation. *)
	  
	  if Queue.length fifoQueues.(ls.rank) > 0 then 
	    begin
	      dump();
	      raise Pending_request
	    end;
	  initOperation request 
	in 
	
	(* Queue may be persistant. We tag operation by their deliver view. *)
	let deliver origin cs msg =
	  (match cs with 
	  | C -> 
	      (match msg with
	      | Fifo (_) -> 
		  fifoOperation.delivered <- succ fifoOperation.delivered
	      | DC (_) -> 
	       (* check_fifo_delivery origin msg; *)
		  dcOperation.delivered <- succ dcOperation.delivered
	      |	_ -> failwith "deliver: bad cast\n");
	      Queue.add msg fifoQueues.(origin)
	  | S -> 
	      (match msg with
	      | Ack (_) -> 
		  Pervasives.incr receivedAck;
		  Queue.add msg fifoQueues.(origin)
	      |	_ -> failwith "deliver: bad send\n") );
	  
       (* check_fifo_sendings (processQueues()) *)
	  processQueues()
	in
	let stable () = not (MI.gt (maxNackedThreshold()) (MI.null())) in 
	let handlers = {
	  dump = dump;
	  deliver = deliver;
	  issue = issue;
	  stable = stable}
	in
	([],handlers)
      in
      
      memberId_map := Name.update_map !memberId_map vs.view;
      if primary && (not !initialised) then 
	begin
	  assert (vs.ltime = 0);	(* we must be a new born singleton *)
	  initialised := true;
	end;
      match vs.xfer_view with 
      | true -> transfer_intf ()
      | false -> running_intf ()

  end
 
