(* No divergence protocol.  

   The protocol considers two kind of operations, those whose metrics
   is zero and the ones whose metrics is above zero.

   The protocol totally orders the execution of operations whose
   metrics is strictly positive. It causally orders (partially) the
   others: null metrics operations are executed after all the previous
   positive metrics operation have been executed. Previous means
   operation that have been executed before an operation.  Also, all
   operations are fifo ordered.

   Althouth we use general term design, this protocol has been devised
   explicitely for the LESS semaphore. In this exemple, TestAndDec
   operations are assigned a strictly positive metrics whereas Inc
   operations are assigned a null metrics. In this context, this
   protocol is proven to guarantee that the local replica state is a
   lower bound of the global state when a TestAndDec operation is to
   be *executed* on its initiation site 

   Last, this protocol does not support view change. So, it works
   currently only for the bench. Support for view change should view
   independant rank

   *)

open Protocol_intf
open Replica_intf
open Measure_intf

module Nodiv =  
  functor (RI : Replica_intf) -> 
    functor (MI : Metrics_intf with type t = RI.metrics) -> struct
      
      open Ensemble
      open View
      open Trans
      open Appl_intf open New

      open Causalts
      open Printf

      type metrics = RI.metrics
      type upcalls = RI.protocol_upcalls
      type replica_op =  RI.op_t
      type acker  = replica_op -> unit

      type seqno = int
      type operationInstance = OperationInstance of replica_op * seqno
      type causalprop = Causalts.s option
      type operationMsg = 
	| Total of operationInstance 
	| Fifo of operationInstance * causalprop
	| NewState of RI.state * (operationMsg Queue.t) array * Endpt.id array

      type operation = operationMsg

 
      type handlers = { 
	  deliver : origin -> cast_or_send -> operation -> (operation naction) list;
	    fetch   : unit -> (operation naction) list; 
	      issue   : (replica_op * acker) -> (operation naction) list; 
		dump    : unit -> unit;
		  stable  : unit -> bool;
	}
	    
      type operationStatus = {
	  mutable initiated : int;
	  mutable delivered : int;
	  mutable scheduled : int;
	} 

       (* The local time stamp generator *)
      let timeStamp = Util.counter ()
	  
      let install (ls,vs) _ upcalls _ _ = 
	let totalOperation = {initiated = 0; delivered = 0; scheduled = 0} in  
	let fifoOperation = {initiated = 0; delivered = 0; scheduled = 0} in  
	
	let alarm =  Elink.alarm_get_hack () in
	let async = Appl.async (vs.group,ls.endpt) in
	let getMyId x = x in
	let string_of_id n = sprintf "[%d:%d]" n (getMyId n) in
	
         (* Vector clock keeping track of the Total->Fifo relationship
	    over execution order  *)
	let siteTime = Causalts.create_state ls.nmembers in

         (* Pending operations *)
	let fifoQueues = Array.init ls.nmembers (fun _ -> Queue.create ()) in 
	let totalQueue = Queue.create () in 

         (* Not acknowledged operations initiated by this member. *)
	let nackedOps = Hashtbl.create 113 in 
	
	let ackOp (OperationInstance (op, ts)) =
	  let (op, localAcker) = Hashtbl.find nackedOps ts in
	  localAcker op;
	  Hashtbl.remove nackedOps ts
	in
	
	let pushOp exec_acker (OperationInstance (op, ts)) =
	  Hashtbl.add nackedOps ts (op, exec_acker);
	in

         (* The core of the delivery algorithm.

	    We iterate forever on fifoqueues until no more operations can be
	    scheduled.  Returns a list of operations to broadcast *)
	let rec processQueues () =
	  let opToSend = ref [] in 
	  let scheduled = ref true in 

	  let scheduleTop q (OperationInstance(op,_)) =
	    match (RI.execute op) with 
	    | true -> 
		Queue.take q;
		scheduled := true; (* keep track of the state change *)
		true
	    | false -> false
	  in
	  
	  while !scheduled do
	    scheduled := false;
	    Array.iteri
	      (fun orig q -> 
		let rec loop_sched q = 
		  if Queue.length q > 0 then 
		    begin
		      let op = Queue.peek q in
		      match op with 
		      | Total(opi) -> 	
			  let (topTotal,origin) = Queue.peek totalQueue in
			  if (orig = origin) && (scheduleTop q opi) then
			    begin
			      assert (topTotal = opi);
			      let _ = Queue.take totalQueue in 
			      Causalts.incr siteTime orig;
			      if (origin = ls.rank) then ackOp opi;
			      totalOperation.scheduled <- succ totalOperation.scheduled;
			      loop_sched q
			    end 
			  else 
			    []
		      | Fifo(opi,None) ->
			  assert (orig = ls.rank);
			  if scheduleTop q opi then 
			    begin
			      ackOp opi;
			      fifoOperation.scheduled <- succ fifoOperation.scheduled;

 			      (* * Only now can we compute the causal dependancies 
				 * and send them to others along with the operation *)
			      (FCast(Fifo(opi, Some(Array.copy siteTime)))) :: (loop_sched q)
			    end
			  else
			    []
		      | Fifo(opi,Some(cs)) ->
			  assert (orig <> ls.rank);
			  if (Causalts.ge siteTime cs) && (scheduleTop q opi) then
			    begin
			      fifoOperation.scheduled <- succ fifoOperation.scheduled;
			      loop_sched q
			    end
			  else 
			    []
		      | NewState _ -> failwith "state initialisation not yet supported" 
		    end
		  else 
		    []
		in
		opToSend := !opToSend @ (loop_sched q) )
	      fifoQueues;	
	  done;
	  let opts = fetch () in
	  match opts with 
	  | [] -> !opToSend
	  | _ -> !opToSend @ opts

	and fetch () =
           (** The core algorithm for initiating operations **)
	  let initOperation (op, exec_acker) = 
	    let muOp = RI.mu op in
	    let opi = OperationInstance (op,  timeStamp ()) in
	    pushOp exec_acker  opi;
	    if MI.gt muOp (MI.null ()) then 
	      begin
		totalOperation.initiated <- succ totalOperation.initiated;
  		[ACast(Total(opi))]		(* Total ordering *)
	      end
	    else
	      begin
		fifoOperation.initiated <- succ fifoOperation.initiated;
		Queue.add (Fifo (opi,None)) fifoQueues.(ls.rank);
		processQueues ()
	      end
	  in
          
          let rec loop = 
            (function 
              | Some(op) -> 
                  (initOperation op) @ (loop (upcalls.RI.getnewop()))
              | None -> [])
          in
          loop (upcalls.RI.getnewop())

	and issue _ = [] in

	let deliver origin cs msg =
	  match cs with 
	  | C -> 
	      (match msg with
	      | Fifo (_) -> 
		  fifoOperation.delivered <- succ fifoOperation.delivered;
		  Queue.add msg fifoQueues.(origin)
	      | Total(opi) -> 
		  totalOperation.delivered <- succ totalOperation.delivered;
		  Queue.add msg fifoQueues.(origin);
		  Queue.add (opi,origin) totalQueue
	      |	NewState _ -> ()) ;
	      processQueues()
	  | _ -> failwith "deliver bad send msg\n"
	in
	
	let dump () =
	  Util.eprintf "siteTime= %s\n" (Causalts.sprint siteTime);
	  Util.eprintf "Total {I= %d, D= %d, S= %d} Fifo {I= %d, D= %d, S= %d}\n"
	    totalOperation.initiated totalOperation.delivered totalOperation.scheduled 
	    fifoOperation.initiated fifoOperation.delivered fifoOperation.scheduled ;
	  Util.eprintf "Total queue (%d)\n" (Queue.length totalQueue);
	  Array.iteri 
	    (fun i q -> 
	      Util.eprintf "Fifo %d (%d), top ts=%s  \n" 
		i (Queue.length q)
		(if (Queue.length q) > 0 then 
		  match Queue.peek q with 
		  | Total (OperationInstance (op,ts)) ->
		      sprintf "%s %d %s" (RI.to_string op) ts "(Total)"
		  | Fifo (OperationInstance (op,ts), _) ->
		      sprintf "%s %d %s" (RI.to_string op) ts "(Fifo)"
		  | NewState _ -> "New state"
		else 
		  "Empty" ))
	    fifoQueues;
	  Util.eprintf "\n"
	in
	let stable () = true in
	let handlers = {
	  dump = dump;
	  deliver = deliver;
	  fetch = fetch;
	  issue = issue;
	  stable = stable
	}
	in
	[],handlers

    end
