(* Epsilon distributed scheduler. *)

open Ensemble
open Util
open View
open Appl_intf open New

open Predicate
(**************************************************************)
let name = Trace.file "BENCH"
let failwith s = failwith (Util.failmsg name s)

let bench_id = ref ""

(**************************************************************)

let size_op = ref 1			(* Fake overhead of semaphore operations *)

open Semaphore
	  
(************* Bench Parametrization **************)
let hbrate  = ref (Time.of_int 1)	(* Ensemble heartbeat rate *)
let oprate  = ref None			(* Issuing rate of operation *)

let no_queue  = ref false		(* queuing mode *)

let primary = ref false  
    
(* The divergence threshold *)
let epsilon = ref 0

(* Bench quorum *)
let nmembers = ref 3  

(* The local divergence threshold *)
let threshold = ref (!epsilon / (!nmembers-1))
    
(* Do we check correctness after execution? *)
let checkOrder = ref false    

(* Do we trace the application queue? *)
let trace_queue = ref false 

(* How many operations do each member initiate *)
let nbOperation = ref 1000

(* How many operations will be executed in the bench *)
let nb_op_exec = ref (!nbOperation * !nmembers)

(**************** Various bookkeeping  ***********************)
(* How many operation do we have issued *)
let issuedOperation = ref 0

(* How many operations do we have initiated *) 
let initiatedOperation = ref 0 

(* How many operation do we have executed *)
let executedOperation = ref 0

(* How many operations do we have initiated AND executed *) 
let initiatedAndExecutedOperation = ref 0 

(* To record randomness of *)
let nbInitiatedTestAndDec = ref 0
let nbInitiatedInc = ref 0

    
(**************** Time features *************************)
open UnixChrono

(* We use three pairs of unix chronos *)
(* chrono 0 measure the time for terminated() to become true
   chrono 1 measure the time for benchTerminated() to become true
   chrono 2 measure the time for readyToQuit() to become true
   *)
let myChronos = UnixChrono.createSet 3
    
(* latencyChrono measure the latency of the execution of operations. *)
(* This is the cumulative time between the initiation of one operation and 
   its local execution *)
let latencyChrono = UnixElapsedChrono.create ()

(* Cumulative latency of operations *) 
(* This is the cumulative time between the issuing of an operation by
the application and its local execution *)
let op_latency = ref (UnixElapsedTimeSet.null())


(* record (reverse order) of the heartbeat occurrence time *)
let heartbeat_queue = ref []

(* record (reverse order) of the response time *)

let response_time = ref []

(* record (reverse order) of the service time *)
let service_time = ref []

let dump_times fname = 
  let dump_log f msg log = 
    if (List.length log) > 0 then 
      begin
	let moments l = 
	  let bi_sum (m1,m2) n = (m1 +. n, m2 +. (n *. n)) in 
	  List.fold_left bi_sum (0.0, 0.0) l
	in
	let n = (List.length log) in 
	let (m1,m2) = moments log in 
	let mean = m1 /. (float n) in
	let var = m2 /. (float n) -. mean *. mean in
	Printf.fprintf f "# %s [n= %d, mean= %f, variance= %f]\n" msg n mean var;
	List.iter (fun t -> Printf.fprintf f "%f\n" t) log;
	Printf.fprintf f "\n\n"; (* New data set -- gnuplot convention *)
      end
  in
  
  let heartbeats =
    let last_h = ref (0.0) in 
    let hb = ref [] in 
    List.iter
      (fun t ->
	hb := (t -. !last_h) :: !hb;
	last_h := t)
      (List.rev !heartbeat_queue);
    let res = 
      match List.rev !hb with 
      | head :: res -> res (* Discard the first (invalid) element *)
      | _ -> []	
    in
    res
  in

  let f = open_out fname in
  dump_log f "Inter heartbeat times" heartbeats;
  dump_log f "Response times" (List.rev !response_time);
  dump_log f "Service times" (List.rev !service_time);
  close_out f;
  Util.eprintf "Heartbeat, response and service times dumped in %s\n" fname


let startChronos () =
  for i = 0 to ((Array.length myChronos) - 1) do
    UnixChrono.start myChronos.(i)
  done

let printChronos () = 
  Array.iteri 
    (fun i chrono -> 
      Util.eprintf "\tstep %d : %s\n" (i+1) 
	(UnixTimeSet.to_string (UnixChrono.getTime chrono)) )
    myChronos

let condActions predicate chrono msg = 
  let action () = 
    if (UnixChrono.getStatus myChronos.(chrono) = UnixChrono.Running) then
      begin
	UnixChrono.stop myChronos.(chrono);
	Util.eprintf "%s" msg;
      end
  in
  Predicate.makeCondAction action predicate

(*************************** Operations Types **************)

(* How nonqueue operations are managed.


   Nonqueue operations are still enqueued.  This avoid to compute the
   (Time.ge <current time> !next_action_time) predicate in the fetch
   path. In the 'fetch' path, this predicate would be evaluated at the
   rate the operations are received by a member. Currently the
   predicate is computed only once at each heartbeat (either regular
   or asynchronous).


   Still there should be no queue effect as an operation is enqueued
   only when there is no pending operations. As fetch is expected to
   be call right after the heartbeat, the only effect should be the
   time of the fetch path inside the protocol. This one is presumably
   very low. *)


let next_action = ref (Some Time.zero)	(* Next operation issuing. Embodies the thinking time *)

module SemaphoreRaw = Semaphore
open Measure_intf


module Replica_upcalls = 
  functor (A : ADT) -> struct 
    type op_t = A.op_t
    type state =  A.state

    let install_state = A.install_state
    let get_state = A.get_state
	
    let execute op = 
      match (A.execute op) with 
      |	true -> incr executedOperation; true
      |	false -> false

    let to_string = A.to_string 
end

open Protocol_intf

(* Replica wrapper *) 
module Replica = 
  functor (A : ADT) -> 
    functor (BH : Behavior with type op_t = A.op_t) -> 
      functor (P : Protocol_intf with type op_t = A.op_t) ->
  struct 
    open Ensemble
    open View
      
    type op_t = P.op_t
    type operation = P.operation
    type acker      = op_t -> unit
    type issuer     = (op_t * acker) -> (operation naction) list

      (* Queue of the operations issued by the application 
	 but still not locally executed *) 
    let (appl_operationsQueue : (op_t * acker) Queue.t)  = Queue.create ()

    let am_coord = ref false
    let pendingOperation = ref None
    let service_start = ref 0.0
    let next_action_time = ref Time.zero	(* Next operation issuing. Embodies the thinking time *)

     (* The behavior may be parameterized by view specific files (coordinator) *)
    let install ((ls,vs) as vf) issue = 
      let getNewOpRaw = BH.install vf !size_op in 
      let generateNewOp () = 
	let newRequest = getNewOpRaw () in
	let executed_acker () =
	  match !oprate with 
	  | Some (rate) ->
	      let t1 = UnixElapsedTimeSet.getTime () in
	      (fun _ -> 
		let t2 = UnixElapsedTimeSet.getTime () in
		let service_t = (t2 -. !service_start) in 
		let response_t = (t2 -. t1) in 
		service_time := service_t :: !service_time;
		response_time := response_t :: !response_time;
		UnixElapsedChrono.stop latencyChrono;
		op_latency := !op_latency +. response_t;
		incr initiatedAndExecutedOperation;
		pendingOperation := None;
		if !no_queue then 
		  begin
		    next_action_time := Time.add (Time.gettimeofday()) rate;
		    next_action := Some !next_action_time
		  end
		    )
	  | None ->
	      (fun _ -> 
		incr initiatedAndExecutedOperation;
		pendingOperation := None)
	in
	incr issuedOperation;
	(newRequest, executed_acker())
      in

      let old_queue_length = ref 0 in 
      let pendingRequest () = 
	match !pendingOperation with 
	| Some(_) -> true
	| None -> false
      in
      
      let getNewOp () =
	let initiate ((op,_) as request) = 
	  incr initiatedOperation;
	  pendingOperation := Some(op);
	  request
	in

	if pendingRequest() then 
	  None
	else
	  match !oprate with
	  | Some (_) ->
	      if Queue.length appl_operationsQueue <= 0 then 
		None 
	      else
		begin
		  service_start := UnixElapsedTimeSet.getTime ();
		  UnixElapsedChrono.resume latencyChrono;
		  Some (initiate (Queue.take appl_operationsQueue))
		end
	  | None ->
	     (* Issuing and Initiating are equivalent when
		not controlled by hearbeat. *)
	      if (!issuedOperation >= !nbOperation) then
		None
	      else
		Some(initiate (generateNewOp ()))
      in

       (* When application issues operations at the oprate rate or
	  when in stalling (no_queue=true) mode (see note above).  In
	  stalling mode, a correctness issue is (i) to enqueue an
	  operation as soon as possible but not too soon (we have no
	  way, currently, to differenciate regular heartbeat from
	  scheduled async callback) and (ii) *not to enqueue* more than
	  one operation.  (Time.ge time !next_action_time) enforce
	  (i). (ii) is achieved by the guard (not (pendingRequest())).
	  *)

      let heartbeat time = 
	let rec loop = function 
          | Some op -> (issue op) @ (loop (getNewOp()))
          | None -> []
	in
	(match !oprate with
	| Some (_) ->
	    if !issuedOperation < !nbOperation then 
	      begin
		if !trace_queue then
		  begin
		    let queue_length = Queue.length appl_operationsQueue in
		    if !old_queue_length <>  queue_length then 
		      begin
			Util.eprintf "%d/" queue_length;
			old_queue_length :=  queue_length;
		      end
		  end;
		if (not !no_queue) || ((not (pendingRequest())) && (Time.ge time !next_action_time)) then 
		  let newRequest = generateNewOp () in
		  Queue.add newRequest appl_operationsQueue
	      end
	| None -> ());
	loop (getNewOp ())
      in
      heartbeat

    let dump () =     
      Util.eprintf "Application queue length= %d\n" (Queue.length appl_operationsQueue)
  end

module RawReplica_upcalls = Replica_upcalls(Semaphore)

(* The Semaphore Access Object, with no Log *)
(* module RawReplica = Replica_up(Semaphore)(RawBehavior) *)

(* The Semaphore Access Object, Instrumented for logging operations *)
(*
module SemaphoreLog = Log.Log (Scalar_metrics)(Semaphore)(RawBehavior)(DC_semaphore_measure)
module CheckedReplica = Replica (SemaphoreLog.ADT)(SemaphoreLog.Measure)(SemaphoreLog.Behavior)
open SemaphoreLog
*)

open Div
module DivSemaphore_notChecked = Div(RawReplica_upcalls)(Scalar_metrics)(DC_semaphore_measure)

module RawReplica = Replica(RawReplica_upcalls)(RawBehavior)(DivSemaphore_notChecked)



let summary () =
  Util.eprintf "\n";
  Util.eprintf "Semaphore state=  %d\n" !Semaphore.theSemaphore;
  Util.eprintf "Operations: Issued= %d, Initiated= %d, I+E= %d, E= %d\n"
    !issuedOperation !initiatedOperation 
    !initiatedAndExecutedOperation !executedOperation;
  Util.eprintf "Execution times (sec):\n";
  printChronos();
  match !oprate with 
  |  Some (_) ->
      if (!nbOperation > 0) && (!initiatedAndExecutedOperation > 0) then 
	begin
	  let latency = UnixElapsedChrono.getTime latencyChrono in
	  Util.eprintf "Latency (Initiation->Execution) (sec): %s (Average=%f,% f)\n"
	    (UnixElapsedTimeSet.to_string latency)
	    (latency /. (float !initiatedAndExecutedOperation))
	    (latency /. (float !nbOperation));
	  Util.eprintf "Latency (Issuing->Execution) (sec): %f (Average=%f, %f)\n"
	    !op_latency
	    (!op_latency /. (float !initiatedAndExecutedOperation))
	    (!op_latency /. (float !nbOperation));

	  let plus a b = a +. b in 
	  let cumulative_response_t =
	    List.fold_left plus 0.0 !response_time in 
	  let cumulative_service_t = 
	    List.fold_left plus 0.0 !service_time in 
	  Util.eprintf "From queues: (Initiation->Execution= %f) (Issuing->Execution= %f)\n"
	    cumulative_service_t cumulative_response_t
	end
      else
	Util.eprintf "No operations still executed -- no latency infos\n"
  | None -> ()


(***********************************************************)
(* The interfaces associated to various phase of the bench *)

let waitingForMoreMember_intf (ls,vs) = 
  let receive origin bk cs msg = [||] in 
  let block () = [||] in
  let heartbeat time' = [||] in 
  let handlers = 
    { flow_block= (fun _ ->());
      receive = receive ;
      block = block ;
      heartbeat = heartbeat ;
      disable = Util.ident } 
  in
  let actions = [||] in
  Util.eprintf "Waiting for more members (%d)\n" !nmembers;
  actions,handlers


let postpone_intf (ls,vs) delay = 
  let trigger_time = Time.add (Time.gettimeofday()) (Time.of_int delay) in
  let prompt = ref false in
  let receive origin bk cs msg = [||] in 
  let block () = [||] in
  let heartbeat = 
    if ls.am_coord then
      (fun time -> 
	if (not !prompt) && (Time.Ord.ge time trigger_time) then
	  begin
	    prompt := true;
	    [|Control(Prompt)|]
	  end
	else
	  [||])
    else
      (fun time -> [||])
  in 
  let handlers = 
    { flow_block= (fun _ ->());
      receive = receive ;
      block = block ;
      heartbeat = heartbeat ;
      disable = Util.ident } 
  in
  let actions = [||] in
  if ls.am_coord then
    Util.eprintf "Postponing bench start for %d sec (master)\n" delay
  else
    Util.eprintf "Postponing bench start for %d sec (slave)\n" delay;
  actions,handlers

let end_intf (ls,vs) = 
  let receive origin bk cs msg = [||] in 
  let block () = [||] in
  let heartbeat time' = [||] in 
  let handlers = 
    { flow_block= (fun _ ->());
      receive = receive ;
      block = block ;
      heartbeat = heartbeat ;
      disable = Util.ident } 
  in
  let actions = [|Control(Leave)|] in
  Util.eprintf "Leaving...\n"; 
  actions,handlers


      
let exit_fun () = 
  Util.eprintf "That's all folks!\n"; 
  exit 0 

let exit_no_log () = 
  summary (); 
  dump_times (!bench_id ^ "times");
  exit_fun ()
      
let exit_log () =
  summary (); 
  dump_times (!bench_id ^ "times")

(*
  SemaphoreLog.run (!bench_id ^ "log") !nmembers !nb_op_exec !epsilon exit_fun
*)

open Protocol_intf
open Replica_intf

module Bench_intf =
  functor (P : Protocol_intf with type metrics = int) -> 
    functor (RI : Replica_intf  with type op_t = P.op_t and type operation = P.operation) -> 
  struct 

   (* This is the interface interacting with the protocol *)
    let running_intf (ls,vs) ri_heartbeat (handlers : P.handlers) = 

     (* We have initiated and executed all our own operations *)
      let terminated () = (!initiatedAndExecutedOperation >= !nbOperation) in 
      
     (* We have terminated the bench *)
      let benchTerminated () = terminated () && (!executedOperation >= !nb_op_exec) in
      
     (* We can quit the bench group when we have terminated AND
	received all bench operations -- remote ones as well as ours. *)
      let readyToQuit () = benchTerminated () && (handlers.P.stable()) in
      
     (* Display the corresponding message and  stop the corresponding chrono
	when the corresponding predicate become true *)
      let check_terminated = 
	condActions terminated 0 "Operations initiated!\n"
      in
      let check_benchTerminated = 
	condActions benchTerminated 1 "All operations executed!\n"
      in
      let check_readyToQuit = 
	condActions readyToQuit 2 "Protocol stable!\n"
      in
      let sendBuffer = ref [] in 

      let receive origin bk cs msg =
	sendBuffer := !sendBuffer @ (handlers.P.deliver origin cs msg);
	let actions = 
	  if bk=U then
	    begin
	      let opts =
		if !no_queue then 
		  begin
		    match !next_action with
		    | Some t -> 

		      (* We request an heartbeat at the end of the
			 thinking period *)

			let act = Control (Timeout t) in 
			next_action := None;
			act :: !sendBuffer 
		    | None -> !sendBuffer
		  end
		else
		  !sendBuffer 
	      in 
	      sendBuffer := [];
	      opts
	    end
	  else
	    []
	in	
	check_terminated();
	check_benchTerminated();
	check_readyToQuit();
	Array.of_list actions
      in 

      let check_dump () = 
	let dump () = handlers.P.dump () in
	let (rc,wc,ec) = Unix.select [Unix.stdin] [] [] 0.0 in 
	if rc <> [] then 
       (* we do have input on stdin *) 
	  begin
         (* fetch the new line *)
	    let command = Pervasives.input_line Pervasives.stdin in
	    Util.eprintf "hbrate= %s, oprate= %s\n" 
	      (Time.to_string !hbrate) 
	      (match !oprate with 
	      | Some(t) -> (Time.to_string t)
	      | None -> "None");
	    summary ();
	    dump (); RI.dump();
	    if ((String.length command) > 0) then 
	      [Control(Dump)]
	    else
	      []
	  end
	else 
	  []
      in
      
      let heartbeat time =
	heartbeat_queue := (UnixElapsedTimeSet.getTime ()) :: !heartbeat_queue;
	if ls.am_coord && (readyToQuit ()) then 
	  [|Control(Prompt)|]
	else
	  let actions = (check_dump()) @ (ri_heartbeat time) in
	  let opts = 
	    if !no_queue then 
	      begin
		match !next_action with 
		| Some t -> 
		 (* We request an asynchronous heartbeat at the end of the thinking period *)
		    let act = Control (Timeout t) in 
		    next_action := None;
		    act :: !sendBuffer 
		| None -> !sendBuffer
	      end
	    else
	      !sendBuffer 
	  in
	  check_terminated();
	  check_benchTerminated();
	  check_readyToQuit();
	  sendBuffer := [];
	  Array.of_list (opts @ actions)
      in

      let handlers = { 
	flow_block= (fun _ -> ());
	block = (fun _ -> handlers.P.dump ();[||]);
	receive = receive;
	heartbeat = heartbeat ;
	disable = (fun () ->()); 
      } 
      in
      let actions = [||] in
      Util.eprintf "Here we are!\n";
      startChronos();
      [||],handlers

(* Manage the various phases of the bench *)
(* * Right now, there is three phases: 
   * First phase is waiting for the quorum of members
   * Second phase is running the bench itself
   * Third phase is sending and analysing execution logs if required
   
   * There can be an optional additional phase 1 bis for postponing the
   * beginning of the bench.
   *)

    let bench_intf exit_fun postpone = 
      let startnow = ref (postpone = None) in 
      let benchExecuted = ref false in 
      let install ((ls,vs) as vf) = 
	if not !benchExecuted then 
	  begin
	    Util.eprintf "view %d %d %s\n"
	      ls.nmembers ls.rank
	      (String.concat " " 
		 (Array.to_list 
		    (Array.map Endpt.string_of_id (Arrayf.to_array vs.view))));

	    if ls.nmembers < !nmembers then	
   	    (* Just wait for more members *)
	      waitingForMoreMember_intf vf
	    else	
	      if not !startnow then
		begin
		  match postpone with 
		  | Some(delay) -> 
		      startnow := true;
		      postpone_intf vf delay
		  | _ -> failwith "inconsistent postpone"
		end
	      else
		begin
	         (* The real work *)
		  let (actions, hdlrs) = P.install (ls,vs) !primary  !nmembers !threshold in
		  let ri_handlers = RI.install vf hdlrs.P.issue in
		  let (_,handlers) = running_intf vf ri_handlers hdlrs in 
		  benchExecuted := true;
		  (Array.of_list actions),handlers
		end
	  end
	else
	  end_intf vf
      in
      Util.eprintf "*hbrate= %s, oprate= %s, threshold= %d\n" 	
	(Time.to_string !hbrate) 
	(match !oprate with 
	| Some(t) -> (Time.to_string t)
	| None -> "None")
	!threshold;
      full { heartbeat_rate      = !hbrate ;
	     install             = install ;
	     exit                = exit_fun }
  end

(* 
open Nodiv
module NodivSemaphore_notChecked = Nodiv(RawReplica)(Scalar_metrics)
module NodivSemaphore_Checked = Nodiv(CheckedReplica)(Scalar_metrics)

module Bench_NoDiv_notChecked = Bench_intf (RawReplica) (NodivSemaphore_notChecked)
module Bench_NoDiv_Checked = Bench_intf(CheckedReplica)(NodivSemaphore_Checked)
*)

(* module DivSemaphore_notChecked = Div(RawReplica)(Scalar_metrics) *)

(* module DivSemaphore_Checked = Div(CheckedReplica)(Scalar_metrics) *)

module Bench_Div_notChecked = Bench_intf (DivSemaphore_notChecked)(RawReplica) 
(* module Bench_Div_Checked = Bench_intf(CheckedReplica)(DivSemaphore_Checked) *)

(**************************************************************)
(* We put the minimal layer for vsync *)
let noDivStack = 
  String.concat ":" 
    [ "Top:Heal:Leave:Inter:Intra:Elect:Merge:Sync:Stable:Vsync";
      "Extended_appl";
      "Fifo";
      "Chk_total_latency";
      "Sequencer_sticky";
      "Collect";
      "Frag";
      "Pt2ptw:Mflow";
      (* "Chk_fifo"; *)
      "Pt2pt:Mnak:Bottom"]

let divStack = 
  String.concat ":" 
    [ "Top:Heal:Leave:Inter:Intra:Elect:Merge:Sync:Stable:Vsync";
      "Top_metric";
      "Chk_div_latency";
      "FifoCS";
      "Frag";
      "Pt2ptw:Mflow";
      (* "Chk_fifo"; *)
      "Pt2pt:Mnak:Bottom"]

let divStack_subcast = 
  String.concat ":" 
    [ "Top:Heal:Leave:Inter:Intra:Elect:Merge:Sync:Stable:Vsync";
      "Top_appl";
      "Applsubcast";
      "Frag:Pt2ptw:Mflow";
      (* "Chk_fifo"; *)
      "Pt2pt:Mnak:Bottom"]


let set_oprate i =
  let t = Time.of_ints 0 (i*1000) in 
  oprate := Some(t);
  if not !no_queue then 
    hbrate := t

let threshold_arg = ref None 
let nb_op_arg = ref None 
let stack_subcast = ref false
let ping_period = ref 0
let postpone= ref None
let sequencer = ref None
let endpt = ref None

open Semaphore
open BehaviorCategory

let bench_parsing = [
  "-n", Arg.Int(fun i -> nmembers := i), "quorum to start the bench" ;
  "-hbrate", Arg.Int(fun i -> hbrate := Time.of_ints 0 (i*1000)), "heartbeat period (msec)" ;
  "-size", Arg.Int(fun i -> size_op := i), "operation overhead (bytes)" ;
  "-oprate", Arg.Int (set_oprate), "op period (msec)";
  "-noqueue", Arg.Unit (fun () -> no_queue := true), "Stall mode";
  "-epsilon", Arg.Int(fun i -> epsilon := i), "epsilon value";
  "-threshold", Arg.Int(fun i -> threshold_arg := Some i), "epsilon value";

  "-subcast", Arg.Unit(fun () -> stack_subcast := true), "use subcast for pt2pt" ;

  "-nbop", Arg.Int(fun i -> nbOperation := i), "nb of operation to initiate" ;
  "-exec", Arg.Int(fun i -> nb_op_arg := Some i), "nb of operation to execute in the system" ;
  "-init", Arg.Int(fun i -> theSemaphore := i), "semaphore init value" ;
  "-check", Arg.Unit(function () -> checkOrder := true), "check for correct ordering at the end of the execution";
  "-traceq", Arg.Unit(function () -> trace_queue := true), "trace the application queue length";
  "-consumer", Arg.Unit(function () -> behavior := Consumer),"set the proxy behavior as a consumer";
  "-random", Arg.Unit(function () -> behavior := Random),"set the proxy behavior as a consumer";
  "-producer", Arg.Unit(function () -> behavior := Producer),"set the proxy behavior as a producer (default)";
  "-conscoord", Arg.Unit(function () -> behavior := CoordinatorIsConsumer),"set the coordinator to be the only consumer";
  "-prodcoord", Arg.Unit(function () -> behavior := CoordinatorIsProducer),"set the coordinator to be the only producer";

  "-bench_id", Arg.String(fun s -> bench_id := s), "Arbitrary string identifying the bench" ;
  "-endpt", Arg.String(fun s -> endpt := Some s), "Arbitrary string uniquely identifying this member (endpt)" ;
  "-sequencer", Arg.String(fun s -> sequencer := Some s), "Arbitrary string identifying the bench" ;

  "-ping", Arg.Int(fun n -> ping_period := n), "Ping period (millisecond)" ;
  "-postpone", Arg.Int(fun n -> postpone := Some n), "Time to wait (seconds) before starting the bench";
]


let default_info name = 
  match !endpt with 
  | Some s -> 
      let modes = Arge.get Arge.modes in
      let properties = Arge.get Arge.properties in
      let key = Arge.get Arge.key in
      let key = match key with
      | None -> Security.NoKey
      | Some s -> Security.Common (Buf.pad_string s)
      in
      let protocol = Property.choose properties in
      Appl.full_info name (Endpt.extern s) false false protocol modes key
  | None -> Appl.default_info name


let run () =
  Arge.parse
    bench_parsing
    (Arge.badarg name)
    "Bench: Semaphore bench program (Georges Brun-Cottan <bruncott@cs.brandeis.edu>)";

   (* Allows non symetrics bench scheme *)
  nb_op_exec := (match !nb_op_arg with |None->(!nbOperation * !nmembers)|Some(v)->v);
  threshold := (match !threshold_arg with | None->(!epsilon/(!nmembers-1)) | Some(t)->t);
  
  (match !sequencer with 
  | Some (hostname) -> Param.default "sequencer_sticky_name" (Param.String (hostname))
  | None -> ());
    
   (* Param.default "fifocs_iq_init" (Param.Int (2 * !nb_op_exec)); *)
   (* Default is not to ping (ping_period = 0) *)
  Param.default "ping_period" (Param.Time (Time.of_ints 0 (1000 * !ping_period)));
  Param.default "ping_log" (Param.String (!bench_id ^ "ping"));
  Param.default "ping_quorum" (Param.Int !nmembers);

  Param.default "chk_total_latency_log" (Param.String (!bench_id ^ "rtt_seq"));
  Param.default "chk_div_latency_log" (Param.String (!bench_id ^ "rtt_div"));

   (* Put new default parameter better suited for high latency and low bandwith *)
(* 
  Param.default "top_sweep" (Param.Time (Time.of_int 10));
  Param.default "merge_sweep" (Param.Time (Time.of_int 5));
  Param.default "stable_sweep" (Param.Time (Time.of_int 5));
  Param.default "suspect_sweep" (Param.Time (Time.of_int 5));
*)
   (* Put new default parameter better suited for high latency and high bandwith *)
  Param.default "pt2pt_sweep" (Param.Time (Time.of_ints 0 (500*1000) ));

  let sched_instance () = 
    let (ls,vs) = default_info name in
    let divergence_control = !threshold > 0 in 
    let proto_id = 
      Proto.id_of_string 
	(if divergence_control then 
	  if !stack_subcast then 
	    divStack_subcast
	  else
	    divStack 
	else
	  noDivStack) 
    in
    let vs = View.set vs [ Vs_proto_id(proto_id) ] in
    let interface = 
      match (divergence_control, !checkOrder) with 
(*       |	(true,true) -> Bench_Div_Checked.bench_intf exit_log !postpone
      |	(false,true) -> Bench_NoDiv_Checked.bench_intf exit_log !postpone *)
      |	(true,false) -> Bench_Div_notChecked.bench_intf exit_no_log !postpone
(*      |	(false,false) -> Bench_NoDiv_notChecked.bench_intf exit_no_log !postpone *)
      |	_ -> failwith "Not supported"
    in 
    Appl.config_new interface (ls,vs);
    Util.eprintf "%s\n" (Proto.string_of_id vs.proto_id);
  in
  
  sched_instance ();
  Appl.main_loop ()
    
  
let _ =
  try 
    run ()
  with 
  | Assert_failure (msg, pos1, pos2) -> 
      Util.eprintf "Assert : %s -- char %d to char %d\n" msg pos1 pos2;
      exit 1
  | Failure (msg) ->
      Util.eprintf "%s\n" msg;
      exit 1
