(* Log management *)

(* At the end of the execution, the coordinator collect the execution
   logs from members and check the divergence protocol correctness
   properties on it. *)

open Measure_intf 
open Semaphore 

module Log = 
  functor (MI : Metrics_intf) -> 
    functor (AD : ADT) -> 
      functor (BH : Behavior with type op_t = AD.op_t) ->
	functor (MS : Measure_intf with type metrics = MI.t and type  op_t = AD.op_t) ->
  struct
    open Printf
    open Ensemble
    open View
    open Trans
    open Util    
    open Appl_intf
    open New
      
    type rank = Trans.rank
    type seqno = int

    type operationInstance = OperationInstance of MS.op_t * seqno * rank
    type op_t = operationInstance
	  
     (* Collect format *)
    type executionLogQueue = operationInstance list
    type operationMsg = Log of rank * executionLogQueue

     (* Log of the operations executed by this member *)
    let (executionLog : (operationInstance Queue.t)) = Queue.create ()
	
    let myrank = ref 0

    let to_string (OperationInstance (op, ts, rank)) = 
      sprintf "[%s:%d:%d]" (AD.to_string op) ts rank

    let tostring = to_string (* Alias to allow access by ADT *)

    let queue_to_list q =
      let l = ref [] in 
      Queue.iter (fun e -> l := e::!l) q;
      List.rev !l

     (* The time stamp generator *)
    let timeStamp = Util.counter ()
	
    module Behavior = struct 
      type op_t = operationInstance
      type view =  BH.view

      let install ((ls,vs) as vf) size_op = 
	let getRawOp = BH.install vf size_op in
	let getNewOp () = OperationInstance (getRawOp(), timeStamp(), ls.rank) in
	myrank := ls.rank;
	getNewOp
    end
	
    module Measure = struct
      type op_t = operationInstance
      type metrics = MS.metrics
      let mu (OperationInstance (op,_,_)) = MS.mu op
    end

    module ADT = struct
      open Ensemble
      open View
      open Trans
      open Util    
	
      type op_t = operationInstance
      type state = AD.state

      let install_state = AD.install_state
      let get_state = AD.get_state

      let execute (OperationInstance (op,_,_) as opi) =
	Queue.add opi executionLog;
	AD.execute op
	  
      let to_string = tostring
    end



    module type Check = sig 
      (* A memo of the Checking module interface *) 
      type exec_log = operationInstance array
      type exec_logs = exec_log array
      type hash_fun = operationInstance -> int
      type checker = int -> int -> MI.t -> hash_fun -> (unit -> unit) -> exec_logs -> bool

      val run : checker
    end


    module Check_OperationSet : Check = struct
      (* Check that every members have executed the same set of operations *)

      type exec_log = operationInstance array
      type exec_logs = exec_log array
      type hash_fun = operationInstance -> int
      type checker = int -> int -> MI.t -> hash_fun -> (unit -> unit) -> exec_logs -> bool


      exception DifferentOperationNumber of (int array)
      exception OperationNotExecuted of int * operationInstance

      let run nmembers nb_op epsilon hash_f dumplogs logArray =
	try 	
	  Util.eprintf "Check_OperationSet...\n";

          (* Does everybody executed the correct number of operations? *)
	  let lengths = Array.init nmembers (fun i -> Array.length logArray.(i)) in
	  Array.iter
	    (fun i -> if i <> nb_op then raise (DifferentOperationNumber lengths))
	    lengths;

          (* And did any member execute the same operations? *)
	  let operationSet = Hashtbl.create 23459 in
           (* Initialise the set with the operations executed by the coordinator... *)
	  Array.iter
	    (fun x -> Hashtbl.add operationSet (hash_f x) x)
	    logArray.(0);
           (* and check they have been executed by the others! *)
	  Array.iteri
	    (fun who l -> 
	      if who <> 0 then 
		Array.iteri 
		  (fun rank x -> 
		    try 
		      let _ = Hashtbl.find operationSet (hash_f x) in 
		      ()
		    with Not_found->
		      raise (OperationNotExecuted (rank,x)))
		  l)
	    logArray;
	    Util.eprintf "Check_OperationSet done\n";
	    true
	with 
	| DifferentOperationNumber (lengths) -> 
	    Util.eprintf "\nCheck_OperationSet:DifferentOperationNumber: (%d) [" nb_op;
	    Array.iter (fun c -> Util.eprintf "%d\t" c) lengths; 
	    Util.eprintf "]\n";
	    false
	| OperationNotExecuted (rank,ropi) -> 
	    Util.eprintf "\nCheck_OperationSet:OperationNotExecuted: %s" (to_string ropi);
	    Util.eprintf "\nnot executed by member %d\n" rank;
	    dumplogs();
	    false
    end

    module Check_Fifo : Check = struct 
      (* Check Fifo property on all logs *)

      type exec_log = operationInstance array
      type exec_logs = exec_log array
      type hash_fun = operationInstance -> int
      type checker = int -> int -> MI.t -> hash_fun -> (unit -> unit) -> exec_logs -> bool
		  
      let run nmembers nb_op epsilon hash_f dumplogs logArray =
	let res = ref true in 
	let check_log_fifo l =
	  let i = ref 0 in
	  let casts = Array.create nmembers 0 in
	  Array.iter
	    (function (OperationInstance (_,ts,rank) as ropi) ->
	      if casts.(rank) <> ts then 
		begin
		  Util.eprintf "\nMisordering: %s" (to_string ropi);
		  Util.eprintf "\texpected ts= %d\n" casts.(rank);
		  res := false
		end;
	      casts.(rank) <- (succ casts.(rank)) )
	    l
	in
	Util.eprintf "Check_Fifo...\n";
	Array.iteri 
	  (fun i l -> 
	    Util.eprintf "\n\t\tchecking log %d..." i;
	    check_log_fifo l;
	    Util.eprintf "done" )
	  logArray;
	Util.eprintf "\nCheck_Fifo done\n";
	!res
    end

    module Check_Metrics  : Check = struct 
      (* Check that the divergence is less than epsilon on every bounded cut.
	 There is exactly one bounded cut (BC) for each operation *)

      type exec_log = operationInstance array
      type exec_logs = exec_log array
      type hash_fun = operationInstance -> int
      type checker = int -> int -> MI.t -> hash_fun -> (unit -> unit) -> exec_logs -> bool
		  
      exception OutofBoundExecution of MI.t * (int array) * (MI.t array)
      exception OperationNotExecuted of int * operationInstance
      exception FoundOp of int * MI.t
	  
      let run nmembers nb_op epsilon hash_f dumplogs logArray =
         (* Returns the couple of n-uplet (index,delta) of the 
	    bounded cut associated to the operation OP *)
	let bounded_cut rop =
	  let bcI = Array.create nmembers 0 in
	  let bcD = Array.create nmembers (MI.null())  in
	  Array.iteri
	    (fun i l ->
	      try
		let m = ref (MI.null ()) in 
		Array.iteri
		  (fun i ropi ->
		    m := MI.add !m (Measure.mu ropi);
		    if (hash_f ropi) = (hash_f rop) then 
		      raise (FoundOp (i,!m)) )
		  l;
		raise (OperationNotExecuted (i,rop))
	      with FoundOp (index,m) ->
		bcI.(i) <- index;
		bcD.(i) <- m )
	    logArray;
	  (bcI,bcD)
	in
	
       	  (* Returns the sup. of the peer to peer divergence of the given bounded cut *)
	let max a b = 
	  if MI.gt a b then 
	    a
	  else b 
	in
	let abs a = 
	  if MI.gt a (MI.null()) then
	    a 
	  else
	    MI.sub (MI.null()) a
	in
	let maxdiv bcD = 
	  let n = nmembers - 1 in
	  let sup = ref (MI.null()) in
	  for i = 0 to n do 
	    for j = (succ i) to n do
	      sup := max !sup (abs (MI.sub bcD.(i) bcD.(j)))
	    done
	  done;
	  !sup
	in
	let sup = ref (MI.null()) in
	try 
	  Util.eprintf "Check_Metrics...\n";
	  Array.iter
	    (fun op ->
	      if MI.gt (Measure.mu op) (MI.null()) then 
		begin
		       (* We test the divergence on bounded cuts associated to
			  non null metrics only. *)
		  let (bcI,bcD) = bounded_cut op in
		  let divergence = maxdiv bcD in 
		  sup := max !sup divergence;
		  if divergence > epsilon then 
		    raise (OutofBoundExecution (divergence,bcI,bcD))
  		end )
	    logArray.(0);		(* We may choose any logs. BC are independant of
					   which log is chosen. *)
	  Util.eprintf "\nMax divergence = %s\n\t" (MI.to_string !sup);
	  Util.eprintf "Check_Metrics done\n";
	  true
	with 
	| OutofBoundExecution (m,bcI,bcD) -> 
	    Util.eprintf "\nCheck_Metrics:OutofBoundExecution: divergence= %s, epsilon= %s, Operation= %s"
	      (MI.to_string m) (MI.to_string epsilon) (to_string logArray.(0).(bcI.(0)));
	    Util.eprintf "\nbcI=[";
	    Array.iter (fun c -> Util.eprintf "%d " c) bcI; 
	    Util.eprintf "]\t bcD=[";
	    Array.iter (fun c -> Util.eprintf "%s " (MI.to_string c)) bcD; 
	    Util.eprintf "]\n";
	    dumplogs();
	    false
    end

    module Check_DCCausality : Check = struct 
       (* Check Causal property on all logs *)

      type exec_log = operationInstance array
      type exec_logs = exec_log array
      type hash_fun = operationInstance -> int
      type checker = int -> int -> MI.t -> hash_fun -> (unit -> unit) -> exec_logs -> bool
	  
      open Causalts 

      type opi = OPI of operationInstance * Causalts.s 

      let run nmembers nb_op epsilon hash_f dumplogs logArray = 
	let res = ref true in
	let operationSet = Hashtbl.create 23459 in
	let build_causality () = 
	  Array.iteri
	    (fun orig l -> 
	      let siteTime = Causalts.create_state nmembers in
	      Array.iter
		(fun (OperationInstance (_,ts,rank) as ropi) ->
		  if (MI.gt (Measure.mu ropi) (MI.null())) then
		    begin
		      if orig = rank then 
			begin
			  Hashtbl.add operationSet (hash_f ropi) (OPI (ropi, Array.copy siteTime))
			end;
		      Causalts.incr siteTime rank;
		    end)
		l)
	    logArray
	in 
	let check_causality () = 
	  Array.iteri
	    (fun orig l -> 
	      let siteTime = Causalts.create_state nmembers in
	      Array.iter
		(fun (OperationInstance (_,ts,rank) as ropi) ->
		  if (MI.gt (Measure.mu ropi) (MI.null())) then
		    begin
		      let OPI(ropi,cs) = Hashtbl.find operationSet (hash_f ropi) in
		      if not (Causalts.ge siteTime cs) then 
			begin
			  Util.eprintf "\nMisordering on %d:%s" orig (to_string ropi);
			  Util.eprintf "\tts expected= %s\t ts got= %s\n"
			    (Causalts.sprint cs) 
			    (Causalts.sprint siteTime);
			  res := false
		  	end;
		      Causalts.incr siteTime rank;
		    end )
		l)
	    logArray
	in	
	Util.eprintf "Check_DCCausality...\n";
	build_causality();
	check_causality();	
	Util.eprintf "Check_DCCausality done\n";	
	
	if not !res then 
	  begin
	    dumplogs()
	  end;
	!res
    end

	
    let default_checklist =
      [ Check_OperationSet.run;
	Check_Fifo.run;
	Check_Metrics.run;
	Check_DCCausality.run ] 

	
    let checklist = ref default_checklist


     (* Members join a new group and send their execution log to the
	coordinator.  When the coordinator did collect the log of
	every participants, the check begins.  *)

    exception CheckAborted

    let run name nmembers nb_op epsilon exit_fun = 
      let log_intf () = 
	let nbReceivedLog = ref 0 in 
	let logLists = Array.create nmembers [] in
	let getAllLog nmembers = (!nbReceivedLog >= nmembers) in 
	let getAllLogs () = getAllLog nmembers in 
	let executionChecked = ref false in 
	let log_delivered = ref false in 
	let i_am_checker = ref false in 

	let deliver origin log = 
	  Util.eprintf "Got execution log from %d\n" origin; 
	  logLists.(origin) <- log;
	  incr nbReceivedLog;
	  getAllLogs ()
	in
	
	let check logdumpname checklist =
	  let logArray = Array.init nmembers (fun i -> Array.of_list logLists.(i)) in
	  let hash_op (OperationInstance(_, ts, rank)) = ts+(rank*nb_op) in
	  let dumpLogs () = 
	    let logtracename = "logtraces" in 
	    let printLogs f =
	      Array.iter 
		(fun l ->
		  Array.iter 
		    (fun op -> fprintf f "%s\n" (to_string op)) 
		    l;
		  fprintf f "*****\n")
		logArray;
	    in
	    let f = open_out logdumpname in 
	    printLogs f;
	    close_out f;
	    Util.eprintf "Logs printed in \"%s\"\n" logdumpname;
	  in
	  
	  try
	    Util.eprintf "\nChecking distributed properties...\n";
	    List.iter
	      (fun check_f ->
		if not (check_f nmembers nb_op epsilon hash_op dumpLogs logArray) then 
		  raise CheckAborted)
	      checklist;
	    Util.eprintf "\nProperties Checked!\n";
	    true
	  with 
	  | CheckAborted -> 
	      Util.eprintf "\nChecking of properties aborted\n";
	      false
	in

	let checking_intf ((ls,vs) as vf) =
	  let receive origin bk cs msg = [||] in 
	  let block () = [||] in
	  let heartbeat time' = [||] in 
	  let handlers = 
	    { flow_block= (fun _ ->());
	      receive = receive ;
	      block = block ;
	      heartbeat = heartbeat ;
	      disable = Util.ident } 
	  in
	  let actions = [|Control(Prompt)|] in 
	  Util.eprintf "Checking execution logs\n";
	  let _ = check "LogTraces" !checklist in
	  executionChecked := true;
	  actions,handlers
	in

	   (* Send/Collect all the logs *)
	let delivering_intf ((ls,vs) as vf) =
	  let receive origin _ _ (Log(rank,log)) =  
	    assert (ls.am_coord);
	    if (deliver rank log) && ls.am_coord then
	      [|Control(Prompt)|]
	    else
	      [||]
	  in 
	  let block () = [||] in
	  let heartbeat _ =  [||] in 
	  let handlers = 
	    { flow_block= (fun _ ->());
	      block = block ;
	      receive = receive ;
	      heartbeat = heartbeat ;
	      disable = Util.ident } 
	  in
	  let log = queue_to_list executionLog in 
	  let actions = 
	    if ls.am_coord then 
	      if (deliver !myrank log) then
		[|Control(Prompt)|]
	      else
		[||]
	    else
	      [|Send([|vs.coord|], Log(!myrank,log));|]
	  in  
	  i_am_checker := ls.am_coord;
	  log_delivered := true;
	  Util.eprintf "Sending execution logs to the coordinator\n";
	  actions,handlers
	in

	let end_intf (ls,vs) = 
	  let receive origin bk cs msg = [||] in 
	  let block () = [||] in
	  let heartbeat time' = [||] in 
	  let handlers = 
	    { flow_block= (fun _ ->());
	      receive = receive ;
	      block = block ;
	      heartbeat = heartbeat ;
	      disable = exit_fun } 
	  in
	  let actions = [|Control(Leave)|] in
	  (* 
	   (* Ensemble bug, singleton members do not receive EExit *)
	  if ls.nmembers <= 1 then 
	    begin
	      exit_fun ()
	    end;
	     *)
	  actions,handlers
	in

	let waitingForMoreMember_intf (ls,vs) = 
	  let receive origin bk cs msg = [||] in 
	  let block () = [||] in
	  let heartbeat time' = [||] in 
	  let handlers = 
	    { flow_block= (fun _ ->());
	      receive = receive ;
	      block = block ;
	      heartbeat = heartbeat ;
	      disable = Util.ident } 
	  in
	  let actions = [||] in
	  actions,handlers
	in
	
	let install ((ls,vs) as vf) =
	  if !executionChecked then 
	    end_intf vf
	  else 
	    if !log_delivered then 
	      if !i_am_checker then 
		checking_intf vf
	      else
		end_intf vf
	    else
	      if (ls.nmembers < nmembers) then 
		waitingForMoreMember_intf vf
	      else	
		delivering_intf vf
	in
	i_am_checker := false;
	full { heartbeat_rate  = Time.of_int (1);
	       install         = install;
	       exit            = exit_fun; }
      in

      let instance () = 
	let (ls,vs) = Appl.default_info name in
	let proto_id = Property.choose Property.vsync in 
	let vs = View.set vs [Vs_proto_id(proto_id)] in
	let interface = log_intf () in 
	Appl.config_new interface (ls,vs);
      in 
      instance ();
      Appl.main_loop ()
	

	
	
  end
