(* Experiments metrics. Netsim communications are not currently
   supported. So experiments must involve IP communications. *)

open Ensemble
open Util
open Arge
open View
open Appl_intf open New

 (* A tiny module to translate members rank in view independant rank *)
(* open Name *)
open Predicate

(**************************************************************)
let name = Trace.file "METRICS-TEST"
let failwith s = failwith (Util.failmsg name s)
(**************************************************************)

(* The replica itself *)
let theSemaphore = ref 1 

type semOperation = TestAndDec | Inc
    
let executeOperation op = 
  (match op with 
    | TestAndDec -> decr theSemaphore
    | Inc -> incr theSemaphore);
  assert (!theSemaphore >= 0)


(* The divergence threshold *)
let epsilon = ref 0


(************* Bench Parametrization **************)

type behavior = 
  | Consumer				(* We generate TestAndDec operations *)
  | Producer				(* We generate Inc operations *)
  | Random				(* We choose randomly TestAndDec/Inc operations *)
  | CoordinatorIsConsumer		(* The group coordinator is a consumer. Others are producers *)
  | CoordinatorIsProducer		(* The group coordinator is a producer. Others are consumers *)

let behavior = ref Producer		(* default behavior *)

let latency  = ref (Time.of_int 0)
let hbrate   = ref (Time.of_int 1)
let oprate   = ref None


(* Nb of members we wait for before starting the bench *)
let nmembers = ref 3          

(* Do we check correctness after execution? *)
let checkOrder = ref false    

(* Do we initiate operations in burst ? *)
let burst = ref false

(* How many operations do each member initiate *)
let nbOperation = ref 1000    

(**************** Various bookkeeping  ***********************)
(* How many operations do we have initiated *) 
let initiatedOperation = ref 0 

(* How many operation do we have executed *)
let executedOperation = ref 0

(* How many operations do we have initiated AND executed *) 
let initiatedAndExecutedOperation = ref 0 

(* How many operations do we have received (from we and others) *) 
let receivedOperation = ref 0 

(* To record randomness of *)
let nbInitiatedTestAndDec = ref 0
let nbInitiatedInc = ref 0

(* Number of acknowledgement received *)
let receivedAck = ref 0

(* Number of acknowledgement sent *)
let ackedOperations = ref 0

(* Number of totally ordered operations executed *)
let executedTotalOperation = ref 0

(**************** Bench Predicates *************************)

(* We have initiated and executed all our own operations *)
let terminated () = 
  (!initiatedAndExecutedOperation >= !nbOperation)
    
(* We have terminated the bench *)
let benchTerminated () = 
  terminated () && (!executedOperation >= !nbOperation * !nmembers)


(* We can quit the bench group when we have terminated AND
   * received all bench operations -- remote ones as well as ours. 
   *)
let readyToQuit () = 
  benchTerminated () && 
  (* (!receivedAck = !nbOperation * (!nmembers-1)) && *)
  (!ackedOperations = !receivedAck)
   (* && (!receivedOperation >= !nbOperation * !nmembers) *)

(**************** Benchmark features *************************)
open UnixChrono

(* We use three pairs of unix chronos *)
(* chrono 0 measure the time for terminated() to become true
   chrono 1 measure the time for benchTerminated() to become true
   chrono 2 measure the time for readyToQuit() to become true
   *)
let myChronos = UnixChrono.createSet 3
    
(* latencyChrono measure the latency of the execution of operations. *)
(* This is the cumulative time between the initiation of one operation and 
   its local execution *)
let latencyChrono = UnixElapsedChrono.create ()

(* Cumulative latency of operations *)
let op_latency = ref (UnixElapsedTimeSet.null())
    
let startChronos () =
  for i = 0 to ((Array.length myChronos) - 1) do
    UnixChrono.start myChronos.(i)
  done

let printChronos () = 
  Array.iteri 
    (fun i chrono -> 
      Util.eprintf "\tstep %d : %s\n" (i+1) 
	(UnixTimeSet.to_string (UnixChrono.getTime chrono)) )
    myChronos
    

let condActions predicate chrono msg = 
  let action () = 
    if (UnixChrono.getStatus myChronos.(chrono) = UnixChrono.Running) then
      begin
	UnixChrono.stop myChronos.(chrono);
	Util.eprintf "%s" msg;
      end
  in
  Predicate.makeCondAction action predicate

(* Display the corresponding message and  stop the corresponding chrono
   when the corresponding predicate become true *)
let check_terminated = 
  condActions terminated 0 "Finished our job. Waiting for other's operations!\n"
    
let check_benchTerminated = 
  condActions benchTerminated 1 "Bench terminated. Waiting for all operations!\n"

let check_readyToQuit = 
  condActions readyToQuit 2 "Quitting...\n"


(*************************** Operations Types **************)

(* The time stamp generator *)
let timeStamp = Util.counter ()

type rank = Trans.rank
type seqno = int
type causalprop = Causalts.s option
type deliverprop = | DTotal | DFifo

type operationInstance = OperationInstance of semOperation * seqno * causalprop
type remoteOperationInstance = RemoteOperationInstance of operationInstance*rank*deliverprop

type queueEvent = Acknowledgement of operationInstance | RemoteOperation of remoteOperationInstance

type executionLogQueue = remoteOperationInstance list

type operationMsg = 
  | Total of operationInstance
  | Fifo of operationInstance
  | Ack of operationInstance
  | Log of executionLogQueue

(*************************** Execution logs and Traces **************)

(* Queue of the operation issued by the application but still not locally executed *) 
let (appl_operationsQueue : (semOperation * (unit -> unit)) Queue.t)  = Queue.create ()

(* Log of the operations initiated by this member *)
let (initiationLog : (operationInstance Queue.t)) = Queue.create ()

(* Log of the operations executed by this member *)
let (executionLog : (remoteOperationInstance Queue.t)) = Queue.create ()

let logInitiation opi = 
  if !checkOrder then 
    Queue.add opi initiationLog

let logExecution ropi = 
  if !checkOrder then 
    Queue.add ropi executionLog

let write_logs () =
  if !checkOrder then 
    ()					(* currently nop *)


let queue_to_list q = 
  let l = ref [] in 
  Queue.iter 
    (fun e -> l := e::!l)
    q;
  List.rev !l


let summary () =
  write_logs ();
  Util.eprintf "\n";
  Util.eprintf "Semaphore state=  %d\n" !theSemaphore;
  Util.eprintf "Operations: I= %d (%d,%d), I+E= %d, E= (%d,%d), R= (%d,%d) A=%d\n"
    !initiatedOperation !nbInitiatedInc !nbInitiatedTestAndDec
    !initiatedAndExecutedOperation 
    !executedOperation !executedTotalOperation
    !receivedOperation !receivedAck 
    !ackedOperations;
  Util.eprintf "Execution times (sec):\n";
  printChronos();
  let latency = UnixElapsedChrono.getTime latencyChrono in 
  Util.eprintf "Total latency (sec): %s (Average=%f) [%f (Average) %f ]  \n"
    (UnixElapsedTimeSet.to_string latency)
    (latency /. (float !nbOperation))
    !op_latency (!op_latency /. (float !initiatedAndExecutedOperation));
  Util.eprintf "Application queue length= %d\n" (Queue.length appl_operationsQueue)


(* Update various bookkeeping infos used by bench predicates *)
let executeLocalOperation ropi myrank = 
  let RemoteOperationInstance(OperationInstance(op,_,_),rank,_) = ropi in 
  executeOperation op;
  incr executedOperation;
  logExecution ropi;
  if myrank = rank then 
    incr initiatedAndExecutedOperation


let traceOp ropi = 
  let RemoteOperationInstance(opi,rank,_) = ropi in 
  let OperationInstance(op,ts,_) = opi in 
  Util.eprintf "[%s]:%d from %d"
    (match op with 
    | TestAndDec -> "Test&Dec"
    | Inc -> "Inc")
    ts rank

let giveMeAnOp ls vs = 
  match !behavior with 
  | Consumer -> TestAndDec
  | Producer -> Inc
  | Random ->
      let n = Random.int 10000 in 
      if n > 5000 then
	TestAndDec 
      else
	Inc
  | CoordinatorIsConsumer ->
      if ls.am_coord then
	TestAndDec
      else
	Inc
  | CoordinatorIsProducer -> 
      if ls.am_coord then
	Inc
      else
	TestAndDec
      

(* wrap previous giveMeAnOp for basic bookkeeping *)
let giveMeAnOp ls vs = 
  let op =  giveMeAnOp ls vs in 
  (match op with 
  |  TestAndDec -> incr nbInitiatedTestAndDec 
  |  Inc -> incr nbInitiatedInc);
  op


(* Application issues operation at the hearbeat rate *)
let appl_heartbeat ls vs = 
  let newRequest = giveMeAnOp ls vs in
  let t1 = UnixElapsedTimeSet.getTime () in
  let acker () = 
    (fun () -> 
      let t2 = UnixElapsedTimeSet.getTime () in
      op_latency := !op_latency +. (UnixElapsedTimeSet.sub t2 t1);
      let _ = Queue.take appl_operationsQueue in 
      Util.eprintf "!";
      ())
  in
  Util.eprintf "(%d)" (Queue.length appl_operationsQueue);
  Queue.add (newRequest, acker()) appl_operationsQueue;
  Queue.peek appl_operationsQueue


(********************** The two Core predicates ******************** *) 

let mu op =
  match op with 
  | TestAndDec -> 1
  | Inc -> 0

let metrics opi = 
  let OperationInstance(op,_,_) = opi in 
  mu op 

let p1 opi = 
  let OperationInstance(op,_,_) = opi in 
  match op with 
  | TestAndDec -> !theSemaphore > 0
  | Inc -> true

(* 	
let p2 opi = 
  let OperationInstance(op,_) = opi in 
  match op with 
  | TestAndDec -> false
  | Inc -> true
*)

(************************ Ordering stuff *****************)

(* Only the coordinator checks for correctness *)
let nbReceivedLog = ref 0

let getAllLog () = (!nbReceivedLog >= !nmembers)

exception DifferentOperationNumber of (int array)
exception DifferentOperationSet of (int array)
exception OutofBoundExecution of int * (int array) * (int array)
exception OperationNotExecuted of int * remoteOperationInstance
exception FoundOp of int*int

open Printf

let checking_properties logLists origin log = 
  Util.eprintf "Got execution log from %d\n" origin; 
  logLists.(origin) <- log;
  incr nbReceivedLog;
  if getAllLog() then 
    begin 
      let logArray = Array.init !nmembers (fun i -> Array.of_list logLists.(i)) in
      
      let hash_ropi ropi = 
	let RemoteOperationInstance(OperationInstance(_,ts,_),rank,_) = ropi in 
	ts+(rank*(!nbOperation*2))
      in
      
      let printLogs f =
	let printLog l = 
	  let printOp ropi = 
	    let RemoteOperationInstance(opi,rank,_) = ropi in
	    fprintf f "%d\t%d\n" (hash_ropi ropi) (metrics opi)
	  in 
	  Array.iter printOp l;
	  fprintf f "*****\n"
	in 
	
	Array.iter printLog logArray;
      in 

      let checkTrace f msg =
	Util.eprintf "\t%s..." msg;
	f ();
	Util.eprintf "done\n"
      in
      
       (* Check that the divergence is less than epsilon on every bounded cut
          There is exactly one bounded cut (BC) for each operation *)
      let check_metrics () = 
         (* Returns the couple of n-uplet (index,delta) of the bounded cut associated to the operation OP *)
	let bounded_cut rop =
	  let bcI = Array.create !nmembers 0 in
	  let bcD = Array.create !nmembers 0 in
	  Array.iteri
	    (fun i l ->
	      try
		let m = ref 0 in 
		Array.iteri
		  (fun i ropi ->
		    let RemoteOperationInstance(o,_,_) = ropi in
		    m := !m + (metrics o);
		    if (hash_ropi ropi) = (hash_ropi rop) then 
		      raise (FoundOp (i,!m)) )
		  l;
		raise (OperationNotExecuted (i,rop))
	      with FoundOp (index,m) ->
		bcI.(i) <- index;
		bcD.(i) <- m )
	    logArray;
	  (bcI,bcD)
	in
	
	 (* Returns the sup. of the peer to peer divergence of the given bounded cut *)
	let maxdiv bcD = 
	  let n = !nmembers - 1 in
	  let sup = ref 0 in
	  for i = 0 to n do 
	    for j = (succ i) to n do
	      sup := max !sup (abs (bcD.(i) - bcD.(j)))
	    done
	  done;
	  !sup
	in
	
	let sup = ref 0 in
	Array.iter 
	  (fun op ->
	    let RemoteOperationInstance(opi,_,_) = op in 
	    if (metrics opi) > 0 then 
	      begin
		 (* We test the divergence on bounded cuts associated to
		    non null metrics only. *)
		let (bcI,bcD) = bounded_cut op in
		let divergence = maxdiv bcD in 
		sup := max !sup divergence;
		if divergence > !epsilon then 
		  raise (OutofBoundExecution (divergence,bcI,bcD))
  	      end )
	  logArray.(0);			(* We may choose any logs. BC are independant of
					   which log is chosen. *)
	Util.eprintf "\nMax divergence = %d\n\t" !sup
      in

       (* Check Causal property on all logs *)
      let check_causal () = 
	let check_log_causal l =
	  let i = ref 0 in
	  let siteTime = Causalts.create_state !nmembers in
	  Array.iter
	    (function ropi ->
	      let RemoteOperationInstance(OperationInstance(op,ts,cs_opt),rank,kind) = ropi in
	      match cs_opt with
	      | Some (cs) ->
		  if not (Causalts.ge siteTime cs rank) then 
		    begin
		      Util.eprintf "\nMisordering:"; traceOp ropi;
		      Util.eprintf "\tts expected= %s\t ts got= %s\n"
			(Causalts.sprint cs) 
			(Causalts.sprint siteTime) 
		    end;
		  if ((mu op) > 0) then 
		    Causalts.incr siteTime rank;
		  ()
	      | None when ((mu op) > 0) -> 
		  Causalts.incr siteTime rank
	      |	_ -> ())
	    l
	in
	Array.iteri 
	  (fun i l -> 
	    Util.eprintf "\n\t\tchecking log %d..." i;
	    check_log_causal l;
	    Util.eprintf "done" )
	  logArray;
	Util.eprintf "\n"
      in

       (* Check Fifo property on all logs *)
      let check_fifo () = 
	let check_log_fifo l =
	  let i = ref 0 in
	  let casts = Array.create !nmembers 0 in
	  Array.iter
	    (function ropi ->
	      let RemoteOperationInstance(OperationInstance(_,ts,_),rank,_) = ropi in 
	      if casts.(rank) <> ts then 
		begin
		  Util.eprintf "\nMisordering:"; traceOp ropi;
		  Util.eprintf "\texpected ts= %d\n" casts.(rank)
		end;
	      casts.(rank) <- (succ casts.(rank)) )
	    l
	in
	Array.iteri 
	  (fun i l -> 
	    Util.eprintf "\n\t\tchecking log %d..." i;
	    check_log_fifo l;
	    Util.eprintf "done" )
	  logArray;
	Util.eprintf "\n"
      in
      
       (* Check that every members have executed the same set of operations *)
      let check_operation_set () =
	let nbOps = Array.length logArray.(0) in
         (* Does the coordinator executed the correct number of operations? *)
	if nbOps <> (!nbOperation * !nmembers) then
	  begin
	    let na = Array.create !nmembers 0 in
	    Array.iteri 
	      (fun i a -> na.(i) <- (Array.length a))
	      logArray;
	    raise (DifferentOperationNumber na)
	  end;
         (* And did the others do the same? *)
	Array.iteri
	  (fun who a -> 
	    if who <> 0 then 
	      if Array.length a <> nbOps then 
		begin
		  let na = Array.create !nmembers 0 in
		  Array.iteri 
		    (fun i a -> na.(i) <- (Array.length a))
		    logArray;
		  raise (DifferentOperationNumber na)
		end)
	  logArray;
         (* And does any member executed the same operations? *)
	let operationSet = Hashtbl.create 23459 in
         (* Initialise with the operation executed by the coordinator... *)
	Array.iter
	  (fun x -> Hashtbl.add operationSet (hash_ropi x) x)
	  logArray.(0);
         (* and check they have been executed by the others! *)
	Array.iteri
	  (fun who l -> 
	    if who <> 0 then 
	      Array.iteri 
		(fun rank x -> 
		  try 
		    let _ = Hashtbl.find operationSet (hash_ropi x) in 
		    ()
		  with Not_found->
		    raise (OperationNotExecuted (rank,x)))
		l)
	  logArray
      in

      try
	Util.eprintf "\nChecking distributed properties...\n";
	checkTrace check_operation_set "checking operations";
	checkTrace check_fifo "checking fifo";
	checkTrace check_causal "checking causal";
	checkTrace check_metrics "checking metrics";
	Util.eprintf "Done!\n";
      with 
      | DifferentOperationNumber (cut) -> 
	  Util.eprintf "\nERROR: members did not execute the same number of operations (%d * %d = %d) ["
	    !nbOperation !nmembers (!nbOperation * !nmembers);
	  Array.iter 
	    (fun c -> Util.eprintf "%d\t" c)
	    cut; 
	  Util.eprintf "]\n"

      |	OutofBoundExecution (m,bcI,bcD) -> 
	  Util.eprintf "\nERROR: execution diverge (%d) more than epsilon (%d) on the following bounded cut:"
	    m !epsilon;
	  traceOp logArray.(0).(bcI.(0)); 
	  Util.eprintf "\nbcI=[";
	  Array.iter (fun c -> Util.eprintf "%d " c) bcI; 
	  Util.eprintf "]\nbcD=[";
	  Array.iter (fun c -> Util.eprintf "%d " c) bcD; 
	  Util.eprintf "]\n";
	  (* Print the full execution logs *)
	  
	  let f = open_out "logtraces" in 
	  printLogs f;
	  close_out f

      |	OperationNotExecuted (rank,ropi) -> 
	  Util.eprintf "\nERROR:";
	  traceOp ropi;
	  Util.eprintf "\nnot executed by member %d\n" rank;
	  
	  let f = open_out "logtraces" in 
	  printLogs f;
	  close_out f
    end


let interface () =
  let flow_block _ = () in 
  let alarm =  Elink.alarm_get_hack () in
  let executionLogSent = ref false in 
  let install (ls,vs) = 
    begin    
      printf "view %d %d %s\n"
	ls.nmembers ls.rank
	(String.concat " " 
	   (Array.to_list 
	      (Array.map Endpt.string_of_id (Arrayf.to_array vs.view))));
      
      if ls.nmembers < !nmembers then
	(* Just wait for more members *)
	begin
	  Util.eprintf "Waiting for more members (%d)\n" !nmembers;
	  let receive origin bk cs msg = [||] in 
	  let block () = [||] in
	  let heartbeat time' = [||] in 
	  let handlers = 
	    { flow_block= flow_block;
	      receive = receive ;
	      block = block ;
	      heartbeat = heartbeat ;
	      disable = Util.ident } 
	  in
	  (* let actions = [|Control(Dump)|] in *)
	  let actions = [||] in
	  actions,handlers

	end
      else
	begin
	  let async = Appl.async (vs.group,ls.endpt) in
	  let getMyId x = x in 
	  let string_of_id n = sprintf "[%d:%d]" n (getMyId n) in

	  let siteTime = Causalts.create_state ls.nmembers in
	  let pendingOperation = ref None in
	  let threshold = ref 0 in 

           (* Not acknowledged operations initiated by this member. *)
	  let nackedOps = Hashtbl.create 113 in 

	   (* For debugging purpose. We trace collected ack'd entries *)
	  let nackedOpsTraced = Hashtbl.create 113 in 
	  
           (* metrics of not acked operation *)
	  let nackedThreshold = Array.create !nmembers 0 in 
	  
           (* Pending remote operations *)
	  let fifoQueues = Array.init !nmembers (fun _ -> Queue.create ()) in 
	  let totalQueue = Queue.create () in 
	  
	  let pendingRequest () = 
	    match !pendingOperation with 
	    | Some(_) -> true
	    | None -> false
	  in 

	  let ackOp origin op =
	    let OperationInstance (_, ts,_) = op in
	    let for_all a =
	      let rec loop i =
		if i < 0 then true
		else if a.(i) then loop (pred i)
		else false
	      in loop (pred (Array.length a))
	    in
	    try 
	      let (op, localAcker, globalAcker, acked) = Hashtbl.find nackedOps ts in
	      acked.(origin) <- true;
	      nackedThreshold.(origin) <- nackedThreshold.(origin) - (metrics op);
	      let expectRemoteAck  = 
		if origin = ls.rank then 
		  localAcker op
		else
		  true
	      in
	      if (not expectRemoteAck) || (for_all acked) then 
		begin
		  globalAcker op;
		  Hashtbl.add nackedOpsTraced ts (op,acked); 
		  Hashtbl.remove nackedOps ts
		end
	    with Not_found ->
	      summary();
	      let ropi = RemoteOperationInstance (op,origin,DFifo) in 
	      traceOp ropi;
	      
	      try 
		let (gcop, acked) = Hashtbl.find nackedOpsTraced ts in 
		let ropi = RemoteOperationInstance (gcop,origin,DFifo) in 
		Util.eprintf "operations already collected!\n";
		traceOp ropi;
		Array.iter
		  (fun x -> Util.eprintf "%s\t" (if x then "Acked" else "false"))
		  acked;
		Util.eprintf "\n"
	      with Not_found -> 
		Util.eprintf "This operation has not been collected\n";
		failwith "ackOp: Acknowledgement of a not referenced operation"
	  in
	  
	  let pushOp localAcker globalAcker opi =
	    let OperationInstance (op, ts,_) = opi in 
	    Hashtbl.add nackedOps ts (opi, localAcker, globalAcker, Array.create !nmembers false);
	    for i = 0 to (!nmembers - 1) do
	      nackedThreshold.(i) <- nackedThreshold.(i) + (mu op)
	    done
	  in
	  
           (* Compute the maximum of threshold amongst every member. *)
	  let maxNackedThreshold () = 
	    let maxThreshold = ref 0 in 
	    Array.iter 
	      (fun t -> if (t > !maxThreshold) then maxThreshold := t)
	      nackedThreshold;
	    !maxThreshold
	  in

	  let opToSendBuf = ref [] in 
	  
	   (* * The core of the delivery algorithm.
	      * We iterate forever on fifoqueues until no operations can be 
	      * scheduled. 
	      *
	      * processQueues may return a list of operation to send and/or broadcast 
	      *)
	  let processQueues () =
	    let opToSend = ref [] in 
	    let scheduled = ref true in 
	    
	    let sendAck opi origin =
	       (* The order is important for FIFO constraint *)
	      incr ackedOperations;
	      opToSend := !opToSend @ [(Send([|origin|], Ack(opi)))]
	    in   

	    let scheduleTop q ropi = 
	      let RemoteOperationInstance(opi,id,_) = ropi in 
	      if (p1 opi) then
		begin
		  Queue.take q;
		  executeLocalOperation ropi (getMyId ls.rank);
		  scheduled := true; (* keep track of the state change *)
		  true
		end
	      else
		false
	    in
	    
	    while !scheduled do
	      scheduled := false;
	      Array.iteri
		(fun orig q -> 
		  let rec loop_sched q = 
		    if Queue.length q > 0 then 
		      begin
			let qe = Queue.peek q in 
			match qe with 
			| Acknowledgement(opi) ->
			    ackOp orig opi;
			    let _ = Queue.take q in
			    scheduled := true; 
			| RemoteOperation(op) ->
			    match op with 
			    | RemoteOperationInstance (opi,id,DFifo) ->
				if orig = ls.rank then 
				  begin
				    if scheduleTop q op then 
				      begin
					ackOp ls.rank opi;
					let OperationInstance(sop,seqno,_) = opi in
 				         (* Only now can we compute the causal dependancies... *)
					let nopi = OperationInstance(sop, seqno, Some(Array.copy siteTime)) in
					if (metrics opi) > 0 then
					  begin
					    Causalts.incr siteTime orig
					  end;
				         (* ... and send them to others along with the operation *) 
					opToSend := !opToSend @ [FCast((Fifo(nopi)))];
					loop_sched q
				      end
				  end
				else
				  begin
				    let OperationInstance(_,_,cs_opt) = opi in
				    (match cs_opt with 
				    |  Some(cs) ->
					if Causalts.ge siteTime cs ls.rank then
					  if scheduleTop q op then
					    begin
					      if (metrics opi) > 0 then 
						begin
						  let _ =  Causalts.incr siteTime orig in
						  sendAck opi id;
						end;
					      loop_sched q
					    end
				    | _ ->
					failwith "No causal dependancies on a fifo operation")
				  end

			    | RemoteOperationInstance (opi,id,DTotal) ->
				let (topTotal,origin) = Queue.peek totalQueue in
				if orig = origin then 
				  begin
				    assert (topTotal =  op);
				    if scheduleTop q op then
				      begin
					Causalts.incr siteTime orig;
					incr executedTotalOperation;
					let _ = Queue.take totalQueue in 
					if (id = ls.rank) then 
					  ackOp ls.rank opi;
					loop_sched q
				      end 
				  end 
		      end
		  in
		  loop_sched q)
		fifoQueues;	
	    done;
	    !opToSend
	  in
	  
	  let getlocalAcker appl_acker withRemoteAck = 
	    UnixElapsedChrono.resume latencyChrono;
	    Util.eprintf "[";
	    (fun op ->
	      UnixElapsedChrono.stop latencyChrono;
	      appl_acker ();
	      Util.eprintf "]";
	      pendingOperation := None;
	      withRemoteAck)
	  in

	  let globalAcker op = () in 
	  
	   (** The core algorithm for initiating operations **)
	  let initOp () =
	    let (newRequest, appl_acker) = appl_heartbeat ls vs in
	    if pendingRequest() then
	       (* Synchronous programming model -> we must wait for the previously 
		  initiated operation to have been executed, before initiating a new one *)
	      []
	    else
	      let muOp = mu newRequest in
	      let threshold = maxNackedThreshold () in 
	      let aboveThreshold m = m > (!epsilon / (ls.nmembers-1)) in

	       (* Here is the metrics criterion *)
	      let initiate () = 
		if aboveThreshold muOp then 
	          (* Total ordering *)
		  (* To Check: do we need to wait the ack of previous operations? *)
		  begin
		    let ts =  timeStamp () in 
		    let opi = OperationInstance (newRequest, ts, None) in
		    Some(Total(opi)) 
		  end
		else
		  if aboveThreshold (threshold + muOp) then 
                   (* We are above the threshold. Wait! *)
		    None
		  else
		    begin
		      let ts =  timeStamp () in 
		      let opi = OperationInstance (newRequest, ts, None) in
		      Some(Fifo(opi))
		    end
	      in
	      let op = initiate () in 
	      match op with 
	      |	Some(Total(opi)) ->
		  incr initiatedOperation;
		  logInitiation opi;
  		  pendingOperation := Some(opi);
		  pushOp (getlocalAcker appl_acker false) globalAcker opi;
		  [ACast(Total(opi))]
	      |	Some(Fifo(opi)) -> 
		  incr initiatedOperation;
		  logInitiation opi;
  		  pendingOperation := Some(opi);
		  pushOp 
		    (if (metrics opi) > 0 then  
		      getlocalAcker appl_acker  true
		    else
		      getlocalAcker appl_acker false)
		    globalAcker opi;

		   (* Fifo operations are not locally bounced. 
		      We must locally enqueued them. *)
		  let ropi = RemoteOperationInstance (opi, getMyId ls.rank, DFifo) in
		  Queue.add (RemoteOperation ropi) fifoQueues.(ls.rank);
		  processQueues ()
	      |	_ -> []
	  in
	  
	   (* Bound the amount of initiated operations according to the test requirements *)
	  let initOperation () = 
	    if (!initiatedOperation < !nbOperation) then
	      let op = initOp () in
	      (match op with 
	      | [] -> () 
	      |	_ -> check_terminated() );
	      op
	    else
	       (* Enough! We did our job! *) 
	      []
	  in
	  
	   (* Initiate and return a list of operations. *) 
	  let initOperations () =
	    let op = initOperation () in 
	    if !burst then
	      async ();
	    op
	  in
	  
	  let quittingP = ref false in 
	  let logLists = Array.create !nmembers [] in

	  let quit () = 
	    if !checkOrder then
	      if not !executionLogSent then
		let log = queue_to_list executionLog in
		executionLogSent := true;
		if ls.am_coord then
		  begin
		    checking_properties logLists ls.rank log;
		    []
		  end
		else
		  begin
		    quittingP := true;
		    [Send([|vs.coord|], Log(log));Control(Prompt)]
		  end
	      else
		if ls.am_coord && (getAllLog ()) then 
		  begin
		    quittingP := true;
		    [Control(Prompt)]
		  end
		else
		  []
	    else
	      begin
		quittingP := true;
		[Control(Prompt)]
	      end
	  in

	  let receiveSend origin msg =
	    match msg with 
	    | Ack(opi) -> 
		incr receivedAck;
		Queue.add (Acknowledgement opi) fifoQueues.(origin);
		(* ackOp origin opi *)
		processQueues ()
	    | Log(log) ->
		assert (ls.am_coord);
		if !checkOrder then 
		  begin
		    checking_properties logLists origin log
		  end;
		[]
	    | _ -> failwith "receiveSend:bad Send\n"
	  in

	  let receiveCast origin msg = 
	    incr receivedOperation;
	    (match msg with 
	    | Total (opi) ->
		let ropi = RemoteOperationInstance(opi, getMyId origin, DTotal) in	
		Queue.add (RemoteOperation ropi) fifoQueues.(origin);
		Queue.add (ropi,origin) totalQueue
	    | Fifo  (opi) ->
		let ropi = RemoteOperationInstance(opi, getMyId origin, DFifo) in
		Queue.add (RemoteOperation ropi) fifoQueues.(origin)
	    | _ -> failwith "receiveCast: bad Cast\n");
	    
	    let opToSend = processQueues () in 
	    check_terminated();
	    check_benchTerminated();
	    check_readyToQuit();
	    opToSend
	  in
	  
	  let receive origin bk cs msg =
	    let opToSend = 
	      match cs with 
	      | S -> receiveSend origin msg
	      | C -> receiveCast origin msg
	    in
	    let actions = 
	      if bk=U then
		begin
	             (* We send possibly pending acknowledgements *)
		  let acts = 
		    if readyToQuit() then 
		      if (not !quittingP) then
			quit()
		      else
			[]
		    else 
		      begin
			match !oprate with 
			| Some(_) -> [] (* We initiate in Heartbeat only *)
			| None -> initOperations ()
		      end
		  in 
		  let opts = !opToSendBuf @ opToSend in 
		  opToSendBuf := [];
		  (opts @ acts) (* We send ack before *)
		end
	      else 
		begin
                   (* we can not send operations now. Save them for latter. *)
		  opToSendBuf := !opToSendBuf @ opToSend;
		  [] 
		end
	    in
	    Array.of_list actions
	  in
	  
	  let block () = [||] in
	  
	  let summaryQueues () =
	    Util.eprintf "siteTime= %s\n" (Causalts.sprint siteTime);
	    Util.eprintf "Total queue (%d)\n" (Queue.length totalQueue);
	    Array.iteri 
	      (fun i q -> 
		Util.eprintf "Fifo %d (%d), top ts=%s  \n" 
		  i (Queue.length q)
		  (if (Queue.length q) > 0 then 
		    (let qe = Queue.peek q in 
		    match qe with 
		    | Acknowledgement (opi) ->
			let OperationInstance(_,_,cs_opt) = opi in
			(match cs_opt with 
			| Some(cs) ->  "Ack " ^ (Causalts.sprint cs)
			| None -> "None" )
		    | RemoteOperation (ropi) ->
			let RemoteOperationInstance(OperationInstance(_,_,cs_opt),_,_) = ropi in
			(match cs_opt with 
			| Some(cs) -> "Operation " ^ (Causalts.sprint cs)
			| None -> "None" ) )
		  else 
		    "Empty" ))
	      fifoQueues;

	    Util.eprintf "Thresholds: ";
	    Array.iter 
	      (fun t -> Util.eprintf "%d\t" t)
	      nackedThreshold;
	    Util.eprintf "\n"
	  in

	  let dump () = 
	    summary ();
	    Util.eprintf "Translation table: myrank= %d, myId= %d\n" 
	      ls.rank (getMyId ls.rank);
	    summaryQueues ()
	  in
	  let check_dump () = 
	    let (rc,wc,ec) = Unix.select [Unix.stdin] [] [] 0.0 in 
	    if rc <> [] then 
              (* we do have input on stdin *) 
	      begin
                 (* fetch the new line *)
		let command = Pervasives.input_line Pervasives.stdin in 
		dump ();
		((String.length command) > 0)
	      end
	    else 
	      false
	  in

	  let heartbeat time' = 
	    let actions = 
	      if check_dump() then 
		[Control(Dump)]
	      else
		if readyToQuit() then
		  if (not !quittingP) then
		    quit()
		  else
		    []
		else
		  initOperations ()
	    in
	    let opts = !opToSendBuf in 
	    opToSendBuf := [];
	    Array.of_list (opts @ actions)
	  in
	  
	  let handlers = {
	    flow_block= flow_block;
	    receive = receive;
	    heartbeat = heartbeat;
	    block = block ;
	    disable = Util.ident
	  }
	  in
	  let actions = 
	     (* I am not sure if I there is some acks to send at this point. *)
	    if readyToQuit() then
	      begin
		check_readyToQuit(); (* it may not have been done at this time *)
		[|Control(Leave)|]
	      end
	    else
	      [||]
	      (* Array.of_list (initOperations ()) *)
	  in 

	  if (not (readyToQuit())) then
	    begin
	      Util.eprintf "Here we are (%d) !\n" vs.coord;
	      startChronos()
	    end
	  else
	    Util.eprintf "Got a View Change\n";
	  actions,handlers
	end
    end
  in
  let exit () = 
    summary (); 
    Util.eprintf "That's all folks...\n"; 
    exit 0 
  in
  full { 
  heartbeat_rate      = !hbrate ;
  install             = install ;
  exit                = exit 
}

(**************************************************************)

let run () =
  Arge.parse 
    [ "-n", Arg.Int(fun i -> nmembers := i), "quorum to start the bench" ;
      "-nbop", Arg.Int(fun i -> nbOperation := i), "nb of operation to initiate" ;
      "-burst", Arg.Unit(fun () -> burst := true), "burst mode" ;
      "-delay", Arg.Int(fun i -> latency := Time.of_ints 0 (i*1000); hbrate := Time.scale 0.5 !latency), "delay to simulate" ;
      "-hbrate", Arg.Int(fun i -> hbrate := Time.of_ints 0 (i*1000)), "heartbeat period (msec)" ;
      "-oprate", Arg.Int (fun i ->
	let t = Time.of_ints 0 (i*1000) in 
	oprate := Some(t);  hbrate := t), "op period (msec)";
      "-init", Arg.Int(fun i -> theSemaphore := i), "semaphore init value" ;
      "-epsilon", Arg.Int(fun i -> epsilon := i), "epsilon value" ;
      "-check", Arg.Unit(function () -> checkOrder := true), "check for correct ordering at the end of the execution";
      "-consumer", Arg.Unit(function () -> behavior := Consumer),"set the proxy behavior as a consumer";
      "-random", Arg.Unit(function () -> behavior := Random),"set the proxy behavior as a consumer";
      "-producer", Arg.Unit(function () -> behavior := Producer),"set the proxy behavior as a producer (default)";
      "-conscoord", Arg.Unit(function () -> behavior := CoordinatorIsConsumer),"set the coordinator to be the only consumer";
      "-prodcoord", Arg.Unit(function () -> behavior := CoordinatorIsProducer),"set the coordinator to be the only producer";
    ]
    (Arge.badarg name)
    "sem: semaphore test program (see Georges Brun-Cottan <bruncott@cs.brandeis.edu>)" ;
  
  let instance () = 
    let (ls,vs) = Appl.default_info name (* (Arge.get Arge.group_name) *) in
    let proto_id = 
      Proto.id_of_string
	(String.concat ":" 
	   [ "Top:Heal:Leave:Inter:Intra:Elect:Merge:Slander:Sync:Stable:Vsync";
	     "Extended_appl";
	      (* "Chk_sync"; *)
  	      (* "Localfifo"; *)
	      (* "Chk_appl_fifo"; *)
	     "Applsubcast";
	     "Fifo";
	     "Sequencer";
	     "Collect";
	     "Frag:Pt2ptw:Mflow";
	      (* "Stlatency"; *)
	     "Chk_fifo";
	     "Pt2pt:Mnak";
	      (* "Dbg_latency"; *)
	     "Bottom" ])
    in
    let vs = View.set vs [Vs_proto_id(proto_id)] in
    let interface = interface () in  
    Param.default "latency_value" (Param.Time !latency);
    Appl.config_new interface (ls,vs);
    Util.eprintf "Pid=%d: %s\n" (Unix.getpid()) (Proto.string_of_id vs.proto_id);
  in
  instance();
  Appl.main_loop ()
  
  
let _ =
  try 
    run ()
  with 
  | Assert_failure (msg, pos1, pos2) -> 
      Util.eprintf "Assert : %s -- char %d to char %d\n" msg pos1 pos2;
      exit 1
  | Failure (msg) ->
      Util.eprintf "%s\n" msg;
      exit 1
