(* The Semaphore Replica module *)
(* Some extra measure modules are provided *)

module Semaphore = struct
  open Printf 
  open Ensemble

  type op_t = TestAndDec of Iovecl.t | Inc of Iovecl.t
  type state = int
    
     (* The replica itself *)
  let theSemaphore = ref 1    
      
  let install_state s = theSemaphore := s 

  let get_state () = !theSemaphore
	
  let execute op = 
    match op with 
    | TestAndDec _ ->
	if !theSemaphore > 0 then 
	  begin
	    decr theSemaphore;
	    true
	  end
	else
	  false
    | Inc _ -> 
	incr theSemaphore;
	true

  let to_string op = 
    sprintf "%s"
      (match op with 
      | TestAndDec _ -> "Test&Dec"
      | Inc _ -> "Inc")
end


module BehaviorCategory = struct
   (* Define the simulated application behavior *)
  type behavior = 
    | Consumer				(* We generate TestAndDec operations *)
    | Producer				(* We generate Inc operations *)
    | Random				(* Randomly TestAndDec/Inc operations *)
    | CoordinatorIsConsumer		(* Coordinator is a consumer. Others are producers *)
    | CoordinatorIsProducer		(* Coordinator is a producer. Others are consumers *)

  let behavior = ref Producer		(* Default behavior *)
end


module type Behavior = sig
  open Ensemble
  open View
    
  type op_t
  type view = View.full

  val install : view -> int -> (unit -> op_t)
end

      
module RawBehavior = struct
  open Semaphore
  open BehaviorCategory
  open Ensemble
  open View
  open Iovecl

  type op_t = Semaphore.op_t
  type view = View.full

  let install ((ls,vs) as vf) size_op = 
    let am_coord = ls.am_coord in 
    let name="Semaphore:RawBehavior" in
    let fast_iov len =
      let len = Buf.ceil len in
      let msg = Iovec.create name len in
      Iovecl.of_iovec name  msg
    in
    let msg = fast_iov size_op in
    let get () =  
      Iovecl.ref name msg;
      msg
    in
    let getNewOp () = 
      match !BehaviorCategory.behavior with 
      | Consumer -> TestAndDec (get())
      | Producer -> Inc (get())
      | Random ->
	  let n = Random.int 10000 in 
	  if n > 5000 then
	    TestAndDec (get())
	  else
	    Inc (get())
      | CoordinatorIsConsumer ->
	  if am_coord then
	    TestAndDec (get())
	  else
	    Inc (get())
      | CoordinatorIsProducer -> 
	  if am_coord then
	    Inc (get())
	  else
	    TestAndDec (get())
    in
    getNewOp
end


open Measure_intf
module DC_semaphore_measure = struct 
  open Semaphore
  open Measure_intf
  open Scalar_metrics
    
  type op_t = Semaphore.op_t
  type metrics = Scalar_metrics.t

  let mu op =
    match op with 
    | TestAndDec _ -> 1
    | Inc _ -> 0
end


(* Not uses right now *)
module Atomic_Semaphore_measure = struct 
  type op_t = Semaphore.op_t
  type metrics = Scalar_metrics.t

  let mu op = 1
end


(* For checking *)
(* 
open Measure_intf
module _ = Semaphore : ADT
module _ = DC_semaphore_measure : Measure_intf
module _ = Atomic_Semaphore_measure : Measure_intf
*)

