(* $Id: chrono.ml,v 1.5 1999/05/28 17:28:23 bruncott Exp $ *)
(* generic chronometer/counter *)

(* Georges Brun-Cottan, Brandeis university, Jan 1999 *)


(* The minimum requirement for a type to be suitable for a chronomoter *)
module type TimeSet = sig
  type time

  val null :  unit -> time
  val getTime : unit -> time
  val add :  time -> time -> time
  val sub :  time -> time -> time
end


(* a simple functor over time functions *)
module Chrono (TS : TimeSet) : 
  sig 
    type time = TS.time
    type chrono
    type chronoStatus = | NeverStarted | Running | Stopped
    type chronoOperation = | Start | Stop | Get | Resume
	
    (* Returns a new fresh chronos array of size (int) *)
    val createSet : int -> chrono array
    val create : unit -> chrono
	
    val start  : chrono -> unit
    val stop   : chrono -> unit
    val resume : chrono -> unit

    val getStatus : chrono -> chronoStatus
	
    (* Returns the elapsed time (excluding stopped period) since the start. *)
    val getTime : chrono -> time

    (* Returns the total elapsed time since the start *)
    val getTimeElapsed : chrono -> time
  end =
  struct 
    type time = TS.time
    type chronoStatus = NeverStarted | Running | Stopped
    type chronoOperation = Start | Stop | Get | Resume
      
    type chrono = 
	{ mutable status     : chronoStatus;
	  mutable start_time : time;
	  mutable stop_time  : time;
	  mutable dead_time  : time; }

    type chronoSet      =  int -> chronoOperation -> chrono
    type chronoInstance =  chronoOperation -> chrono

    let initChrono chrono =
      {	status = NeverStarted; 
	start_time = TS.getTime(); 
	stop_time  = TS.getTime();
	dead_time  = TS.null();
      }

    (* Returns a new fresh chronos set *)
    let createSet nbChrono = 
      Array.init nbChrono initChrono

    let create = initChrono
	
    let start chrono = 
      chrono.start_time <- TS.getTime();
      chrono.status <- Running

    let stop chrono = 
      if (chrono.status <> Stopped) then 
	begin
	  chrono.stop_time <- TS.getTime();
	  chrono.status <- Stopped;
	  if (chrono.status = NeverStarted) then
	      chrono.start_time <- chrono.stop_time
	end
	  
    let resume chrono = 
      (match (chrono.status) with 
      |	Stopped -> 
	  chrono.dead_time <- 
	     TS.add chrono.dead_time (TS.sub (TS.getTime ()) chrono.stop_time)
      |	NeverStarted -> 
	  chrono.start_time <- TS.getTime()
      |	_ -> () );
      chrono.status <- Running
      
    let getStatus theChrono = theChrono.status

    (* Returns the time elapsed since the start *)
    let getTime theChrono = 
      (* assert (theChrono.status <> NeverStarted); *)
      TS.sub
	(if theChrono.status <> Stopped then 
	  TS.getTime()
	else
	  theChrono.stop_time)
	(TS.add theChrono.dead_time theChrono.start_time)
	

    let getTimeElapsed theChrono = 
      assert (theChrono.status <> NeverStarted);
      TS.sub
	(if theChrono.status <> Stopped then 
	  TS.getTime()
	else
	  theChrono.stop_time)
	theChrono.start_time
  end
